package ankistream;

import java.io.File;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.IOException;
import java.awt.Toolkit;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JLabel;
import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem; 
import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;
import net.n3.nanoxml.IXMLParser;
import net.n3.nanoxml.IXMLReader;
import net.n3.nanoxml.XMLWriter;
import net.n3.nanoxml.XMLParserFactory;
import net.n3.nanoxml.StdXMLReader;

/** Represents the top level tag in a anki box xml file and provides methods
  * for reading and writing anki box data to and from a file.
  * <br> 
  * <br> Part of the Anki Stream project, a tool for long-term flash card
  * review.  Anki Stream is distributed under the MIT/X11 free software 
  * license, a copy of which should have come with this distribution.
  * @author Bishop Wilkins
  * @version 0.1
*/ 
public class AnkiBoxData {

	private IXMLElement xml; // The <ankiboxdata> tag.

	// The constituents of the ankiboxdata:
	private Preferences prefs;
	private Box box;
	private CardSet cardset;
	
	/** Creates a new AnkiBoxData object with the given tag as a basis.
	 * @param newXml the ankiboxdata tag to base this data object on.
	 */
	public AnkiBoxData( IXMLElement newXml ) throws RuntimeException {
		xml = newXml;
		prefs = null;
		box = null;
		cardset = null;
		IXMLElement pTag = xml.getFirstChildNamed("preferences");
		IXMLElement bTag = xml.getFirstChildNamed("box");
		IXMLElement cTag = xml.getFirstChildNamed("cardset");
		if ( pTag != null ) { prefs = new Preferences( pTag ); } 
		if ( bTag != null ) { box = new Box( bTag ); } 
		if ( cTag != null ) { cardset = new CardSet( cTag ); } 		
	}
	
	/** Creates a new anki box data object with no preferences, box, or
	  * card set.
	 */
	public AnkiBoxData( ) {
		xml = new XMLElement( "ankiboxdata" );
		prefs = null;
		box = null;
		cardset = null;
	}
		
	/** Loads the first ankiboxdata tag found in the given file.
	  * @param filename name of the file to search for ankiboxdata tag in.
	  * @return a AnkiBoxData object representing the xml data found, or
	  * null if no ankiboxdata tag was found or there was some other
	  * problem with reading the file.
	 */
	public static AnkiBoxData loadDataFromFile( String filename ) {
		
		AnkiBoxData output;

		IXMLParser parser;
		try {
			parser = XMLParserFactory.createDefaultXMLParser();
		} catch ( ClassNotFoundException cnfe ) {
			return null;
		} catch ( IllegalAccessException iae ) {
			return null;
		} catch ( InstantiationException ie ) {
			return null;
		}
		
		InputStreamReader isr;
		IXMLReader reader;
		IXMLElement readxml;
		try {
			FileInputStream fis = new FileInputStream( filename );
			isr = new InputStreamReader( fis, "UNICODE" );
			reader = new StdXMLReader( isr );
			parser.setReader(reader);
			readxml = (IXMLElement) parser.parse();
		} catch ( Exception e ) {
			return null;
		}
		
		if ( readxml.getFullName( ).compareToIgnoreCase( "ankiboxdata" ) 
				!= 0 ) return null;
		try {
			output = new AnkiBoxData( readxml );
		} catch ( RuntimeException re ) {
			return null;
		}
		
		return output;
						
	} // end loadDataFromFile

	public IXMLElement getPreferencesTag( ) {
		return xml.getFirstChildNamed( "preferences" );
	}

	public IXMLElement getBoxTag( ) {
		return xml.getFirstChildNamed( "box" );
	}

	public IXMLElement getCardSetTag( ) {
		return xml.getFirstChildNamed( "cardset" );
	}
	
	public Preferences getPreferences( ) { return prefs; }

	public Box getBox( ) { return box; }

	public CardSet getCardSet( ) { return cardset; }
	
	public void setPreferences( Preferences newprefs ) {
		IXMLElement oldprefs = xml.getFirstChildNamed( "preferences" );
		if ( oldprefs != null ) { xml.removeChild( oldprefs ); }
		xml.addChild( newprefs.toXML( ) ); 
		prefs = newprefs;
	}

	public void setBox( Box newbox ) {
		IXMLElement oldbox = xml.getFirstChildNamed( "box" );
		if ( oldbox != null ) { xml.removeChild( oldbox ); }
		xml.addChild( newbox.toXML( ) );
		box = newbox;
	}

	public void setCardSet( CardSet newset ) {
		IXMLElement oldset = xml.getFirstChildNamed( "cardset" );
		if ( oldset != null ) { xml.removeChild( oldset ); }
		xml.addChild( newset.toXML( ) );
		cardset = newset;
	}
	
	/** Saves the given object (must be either a Preferences, Box, or CardSet)
	  * to the given file. The tag will be saved as a child of the given
	  * file's ankiboxdata tag, replacing any tag with the same full name
	  * that already existed as a child of ankiboxdata.
	  * @param tag the tag to save.
	  * @param file the xml file to save to.
	  * @return true if the tag was saved successfully; false otherwise.
	 */
	public static boolean saveTagToFile( IXMLElement tag, File file ) {

		String fullname = tag.getFullName( );
				
		AnkiBoxData lbd = loadDataFromFile( file.getAbsolutePath( ) );
		if ( lbd == null ) { lbd = new AnkiBoxData(); }
		
		if ( fullname.compareToIgnoreCase( "preferences" ) == 0 ) {
			lbd.setPreferences( new Preferences ( tag ) );
		} else if ( fullname.compareToIgnoreCase( "box" ) == 0 ) {
			lbd.setBox( new Box( tag ) );
		} else if ( fullname.compareToIgnoreCase( "cardset" ) == 0 ) {
			lbd.setCardSet( new CardSet( tag ) );
		}

		OutputStreamWriter osw;
		try {
			FileOutputStream fos = new FileOutputStream( file );
			osw = new OutputStreamWriter( fos, "UNICODE" );
		} catch ( Exception e ) {
			return false;
		}
				
		XMLWriter writer
			= new XMLWriter( osw );
		try {
			osw.write( "<?xml version=\"1.0\"?>\r\n\r\n" );
			osw.write(
				"<!DOCTYPE ankiboxdata SYSTEM \"ankiboxdata.dtd\">");
			osw.write( "\r\n\r\n" );
			writer.write(lbd.xml, true);
			osw.close( );
		} catch ( IOException ioe ) {
			return false;
		}
				
		return true;
	}
		
} // end class AnkiBoxData
