package ankistream;

import java.io.File;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.IOException;
import java.awt.Toolkit;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JDialog;
import javax.swing.JScrollPane;
import javax.swing.JLabel;
import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem; 
import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;
import net.n3.nanoxml.IXMLParser;
import net.n3.nanoxml.IXMLReader;
import net.n3.nanoxml.XMLWriter;
import net.n3.nanoxml.XMLParserFactory;
import net.n3.nanoxml.StdXMLReader;

/** Represents the top level tag in a anki stream xml file and provides methods
  * for reading and writing anki stream data to and from a file.
  * <br> 
  * <br> Part of the Anki Stream project, a tool for long-term flash card
  * review.  Anki Stream is distributed under the MIT/X11 free software 
  * license, a copy of which should have come with this distribution.
  * @author Bishop Wilkins
  * @version 0.1
*/ 
public class AnkiStreamData {

	private IXMLElement xml; // The <ankistreamdata> tag.

	// The constituents of the ankistreamdata:
	private Preferences prefs;
	private CardSet cardset;
	private Stream stream;
	private Box box;
	
	/** Creates a new AnkiStreamData object with the given tag as a basis.
	 * @param newXml the ankistreamdata tag to base this data object on.
	 */
	public AnkiStreamData( IXMLElement newXml ) throws RuntimeException {
		xml = newXml;
		prefs = null;
		cardset = null;
		stream = null;
		box = null;
		IXMLElement pTag = xml.getFirstChildNamed("preferences");
		IXMLElement cTag = xml.getFirstChildNamed("cardset");
		IXMLElement sTag = xml.getFirstChildNamed("stream");
		IXMLElement bTag = xml.getFirstChildNamed("box");
		if ( pTag != null ) { prefs = new Preferences( pTag ); } 
		if ( cTag != null ) { cardset = new CardSet( cTag ); } 		
		if ( sTag != null ) { stream = new Stream( sTag ); } 
		if ( bTag != null ) { box = new Box( bTag ); } 
	}
	
	/** Creates a new anki box data object with no preferences, stream, or
	  * card set.
	 */
	public AnkiStreamData( ) {
		xml = new XMLElement( "ankistreamdata" );
		prefs = null;
		cardset = null;
		stream = null;
		box = null;
	}
		
	/** Loads the first ankistreamdata tag found in the given file.
	  * @param filename name of the file to search for ankistreamdata tag in.
	  * @return a AnkiStreamData object representing the xml data found, or
	  * null if no ankistreamdata tag was found or there was some other
	  * problem with reading the file.
	 */
	public static AnkiStreamData loadDataFromFile( String filename ) {

		AnkiStreamData output;

		IXMLParser parser;
		try {
			parser = XMLParserFactory.createDefaultXMLParser();
		} catch ( ClassNotFoundException cnfe ) {
			return null;
		} catch ( IllegalAccessException iae ) {
			return null;
		} catch ( InstantiationException ie ) {
			return null;
		}
		
		InputStreamReader isr;
		IXMLReader reader;
		IXMLElement readxml;
		try {						
			// Read the file.
			FileInputStream fis = new FileInputStream( filename );
			isr = new InputStreamReader( fis, "UNICODE" );
			reader = new StdXMLReader( isr );
			parser.setReader(reader);
			readxml = (IXMLElement) parser.parse();
		} catch ( Exception e ) {
			return null;
		}
		
		if ( (readxml == null) ||
			 (readxml.getFullName( ).compareToIgnoreCase( "ankistreamdata" ) 
				!= 0) ) return null;
		try {
			output = new AnkiStreamData( readxml );
		} catch ( RuntimeException re ) {
			return null;
		}
		
		return output;
						
	} // end loadDataFromFile

	public IXMLElement getPreferencesTag( ) {
		return xml.getFirstChildNamed( "preferences" );
	}

	public IXMLElement getCardSetTag( ) {
		return xml.getFirstChildNamed( "cardset" );
	}
		
	public IXMLElement getStreamTag( ) {
		return xml.getFirstChildNamed( "stream" );
	}

	public IXMLElement getBoxTag( ) {
		return xml.getFirstChildNamed( "box" );
	}
	
	public Preferences getPreferences( ) { return prefs; }

	public CardSet getCardSet( ) { return cardset; }

	/** Returns a Stream if this dataset has one, otherwise a Box.
	 */
	public CardContainer getCardContainer() {
		if ( stream != null ) { return getStream( ); }
		return getBox( );
	}
	
	public Stream getStream( ) { return stream; }

	public Box getBox( ) { return box; }

	public void setPreferences( Preferences newprefs ) {
		IXMLElement oldprefs = xml.getFirstChildNamed( "preferences" );
		if ( oldprefs != null ) { xml.removeChild( oldprefs ); }
		prefs = newprefs;
		if ( newprefs == null ) { return; }
		xml.addChild( newprefs.toXML( ) ); 
	}

	public void setCardSet( CardSet newset ) {
		IXMLElement oldset = xml.getFirstChildNamed( "cardset" );
		if ( oldset != null ) { xml.removeChild( oldset ); }
		cardset = newset;
		if ( newset == null ) { return; }
		xml.addChild( newset.toXML( ) );
	}

	public void setCardContainer( CardContainer newcc ) {
		if ( newcc instanceof Stream ) {
			setStream( (Stream)newcc );
		} else { // It's a box.
			setBox( (Box)newcc );
		}
	}
	
	public void setStream( Stream newstream ) {
		// TODO: Once in a while removeChild is getting a null pointer exception
		// in nanoxml I don't understand.  Should investigate further then
		// remove the workaround below.
		int tries = 10;
		while ((xml.getFirstChildNamed( "stream" ) != null) & (tries>=0)) {
			try {
				xml.removeChild( xml.getFirstChildNamed( "stream" ) );
			} catch ( NullPointerException npe ) {
				try { Thread.sleep(500); } catch (InterruptedException ie) {}
			}
			tries--;
		}
		stream = newstream;
		if ( newstream == null ) { return; }
		xml.addChild( newstream.toXML( ) );
	}

	public void setBox( Box newbox ) {
		IXMLElement oldbox = xml.getFirstChildNamed( "box" );
		if ( oldbox != null ) { xml.removeChild( oldbox ); }
		box = newbox;
		if ( newbox == null ) { return; }
		xml.addChild( newbox.toXML( ) );
	}

	/** Saves the given object (must be either a Preferences, CardSet, Stream,
	  * or Box) to the given file. The tag will be saved as a child of the given
	  * file's ankistreamdata tag, replacing any tag with the same full name
	  * that already existed as a child of ankistreamdata.
	  * @param tag the tag to save.
	  * @param file the xml file to save to.
	  * @return true if the tag was saved successfully; false otherwise.
	 */
	public static boolean saveTagToFile( IXMLElement tag, File file ) {

		String fullname = tag.getFullName( );
				
		AnkiStreamData lbd = loadDataFromFile( file.getAbsolutePath( ) );
		if ( lbd == null ) { lbd = new AnkiStreamData(); }
		
		if ( fullname.compareToIgnoreCase( "preferences" ) == 0 ) {
			lbd.setPreferences( new Preferences ( tag ) );
		} else if ( fullname.compareToIgnoreCase( "stream" ) == 0 ) {
			lbd.setStream( new Stream( tag ) );
		} else if ( fullname.compareToIgnoreCase( "cardset" ) == 0 ) {
			lbd.setCardSet( new CardSet( tag ) );
		} else if ( fullname.compareToIgnoreCase( "box" ) == 0 ) {
			lbd.setBox( new Box( tag ) );
		} 

		OutputStreamWriter osw;
		try {
			FileOutputStream fos = new FileOutputStream( file );
			osw = new OutputStreamWriter( fos, "UNICODE" );
		} catch ( Exception e ) {
			return false;
		}
				
		XMLWriter writer
			= new XMLWriter( osw );
		try {
			osw.write( "<?xml version=\"1.0\"?>\r\n\r\n" );
			osw.write(
				"<!DOCTYPE ankistreamdata SYSTEM \"ankistreamdata.dtd\">");
			osw.write( "\r\n\r\n" );
			writer.write(lbd.xml, true);
			osw.close( );
		} catch ( IOException ioe ) {
			return false;
		}
				
		return true;
	}
		
} // end class AnkiStreamData
