package ankistream;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Vector;
import java.util.TreeSet;
import java.util.Collections;
import java.util.Random;
import java.util.NoSuchElementException;
import java.util.zip.DataFormatException;
import java.io.StringReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;
import javax.swing.JOptionPane;
import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;

/** A cardboard box, open at the top, separated into partitions into which flash
  * cards can be placed for review, and a record of which cards (by ID number) 
  * are in which partitions.  The <code>Box</code> class also keeps track of a
  * pile of "pre-box" cards for immediate review, a pile of "post-box" cards for
  * long term review. 
  * <br>In addition to the box partitions, there are "waiting piles" next to
  * each partition where cards go immediately after being succesfully reviewed.
  * <br>All these different locations taken together make up the "compartments"
  * of the <code>Box</code> container.  <code>Box</code> method parameters and
  * return values use partition numbers wherever possible, but for non-partition
  * locations, the compartment IDs are used instead.  The meanings of these are
  * as follows:
  * <br>Compartment 0: The pre-box waiting pile.
  * <br>Compartment 1: The pre-box pile.
  * <br>Compartment (n*2)+2: The waiting pile for partition n.
  * <br>Compartment (n*2)+3: Partition n.
  * <br>Compartment (N*2)+2: The post-box pile. (Where N is the total number 
  * of partitions.)
  * <br>Compartment (N*2)+3: A chunk of the post-box currently under review.<br>
  * <br> 
  * <br> Part of the Anki Stream project, a tool for long-term flash card
  * review.  Anki Stream is distributed under the MIT/X11 free software 
  * license, a copy of which should have come with this distribution.
  * @author Bishop Wilkins
  * @version 0.1
*/ 
public class Box implements CardContainer {
	
	private UndoData undodata;
	private ArrayList orderedreviewsessions;
	
	private IXMLElement xml;	// The XML tree with the box data.	
	private Random random;		// For shuffling cards.
	
	/** Creates a new Box using the data in the given XML tag tree.
	  * @param newXML the root ("box" tag) of an XML tree containing the box
	  *  data for this Box. 
	 */
	public Box( IXMLElement newXML ) {
		
		undodata = null;
		
		xml = newXML;
		random = new Random( System.currentTimeMillis( ) );
		
		// Organize the review sessions.
		orderedreviewsessions = new ArrayList();
		IXMLElement tag = xml.getFirstChildNamed( "reviewstatistics" );
		
		if ( tag != null ) {  
			Vector sessions = tag.getChildrenNamed( "reviewsession" );
			Iterator i = sessions.iterator();
			while ( i.hasNext() ) {
				orderedreviewsessions.add( 
					new ReviewSessionReference( (IXMLElement)i.next() ) );
			}
			Collections.sort( orderedreviewsessions );
		}
	}
	
	
	public Box( String cardSetFile ) {
		this( cardSetFile, new int[] { 35, 75, 180, 300, 560 } );
	}
	/** Creates a new Box with default attributes.
	 */
	public Box( String cardSetFile, int[] capacities ) {
		
		undodata = null;
		
		xml = new XMLElement( "box" );
		xml.setAttribute( "name", "New Box" );
		xml.setAttribute( "cardsetfile", AnkiStreamFrame.defaultDataFile );
		xml.setAttribute( "currentpartition", "-2" );
		xml.setAttribute( "right", "0" );
		xml.setAttribute( "wrong", "0" );
		xml.setAttribute(	"secondspercard", 
							String.valueOf(DEFAULT_SECONDS_PER_CARD) );

		IXMLElement newTag = new XMLElement( "prebox" );
		newTag.addChild( new XMLElement( "waiting" ) );
		xml.addChild( newTag );
		for ( int c = 0; c < capacities.length; c++ ) {
			newTag = new XMLElement( "partition" );
			newTag.setAttribute( "capacity", String.valueOf( capacities[c] ) );
			newTag.addChild( new XMLElement( "waiting" ) );
			xml.addChild( newTag );
		}
		xml.addChild( new XMLElement( "postbox" ) );
		xml.addChild( new XMLElement( "postboxreviewpile" ) );
		xml.addChild( new XMLElement( "reviewstatistics" ) );
		
		orderedreviewsessions = new ArrayList();
		random = new Random( System.currentTimeMillis( ) );
	}
	
	/** Gets the root of an XML tree containing the data for this box.
	  * @return the root of an XML tree containing the data for this box.
	 */
	public IXMLElement toXML( ) { return xml; }
	
	private Vector getPartitions( ) {
		return xml.getChildrenNamed( "partition" );
	}	
	
	/** Gets the name of this box.
	  * @return the name of this box
	 */
	public String getName( ) {
		return xml.getAttribute( "name", "Unnamed Box" );
	}
	
	/** Sets the name of this box.
	 */
	public void setName( String newname ) {
		if ( newname == null || newname.length( ) == 0 ) return;
		xml.setAttribute( "name", newname );
	}
	
	/** Repartitions this box based on the given array of partition sizes.
	  * If cards are left without partitions, they are moved to the furthest
	  * back partition.
	 */
	public void rePartition( int[]psizes ) {
		if (psizes == null || psizes.length < 1 ) return;
		if ( psizes.length < getPartitionCount() ) {
			// Move cards without partitions into the highest partition.
			int source = psizes.length;
			int dest = partitionNumberToCompartmentNumber(psizes.length - 1);
			int sourcemax = getPartitionCount();
			for ( ; source < sourcemax; source++ ) {
				moveAll( partitionNumberToCompartmentNumber(source), dest);
			}
		}
		
		// Resize existing partitions.
		Vector parts = getPartitions();
		Iterator i = parts.iterator();
		int c = 0;
		IXMLElement curpart;
		while ( i.hasNext() ) {
			if ( c >= psizes.length ) break;
			curpart = (IXMLElement)i.next();
			curpart.setAttribute( "capacity", String.valueOf(psizes[c]) );
			c++;
		}

		// Add new partitions.
		while ( c < psizes.length ) {
			XMLElement newpart = new XMLElement( "partition" );
			newpart.setAttribute( "capacity", String.valueOf(psizes[c]) );
			xml.addChild( newpart );
			c++;
		}

		// Delete unnecessary partitions.
		while ( i.hasNext() ) {
			curpart = (IXMLElement)i.next();
			xml.removeChild( curpart );
		}
	}

	/** Gets the name of the data file containing the card set for this box.
	 */
	public String getCardSetFile( ) {
		return xml.getAttribute( "cardsetfile", 
			AnkiStreamFrame.defaultDataFile );
	}

	/** Sets the name of the data file containing the card set for this box.
	 */
	public void setCardSetFile( String newfile ) {
		if ( newfile == null || newfile.length( ) == 0 ) return;
		xml.setAttribute( "cardsetfile", newfile );
	}

	/** Gets the seconds per card that this card container uses for reviews.
	 */
	public int getSecondsPerCard( ) {
		return xml.getAttribute("secondspercard",DEFAULT_SECONDS_PER_CARD);
	}

	/** Sets the seconds per card that this card container uses for reviews.
	 */
	public void setSecondsPerCard( int newval ) {
		if (newval < 0) {newval = 0;}
		xml.setAttribute("secondspercard", String.valueOf(newval));
	}

	/** Gets the font size that this card container uses for review questions.
	 */
	public float getQuestionFontSize( ) {
		return Float.parseFloat( xml.getAttribute( "questionfontsize",
				String.valueOf(FontController.getUnicodeFont().getSize()) ) );		
	}

	/** Sets the font size that this card container uses for review questions.
	 */
	public void setQuestionFontSize( int newval ) {
		xml.setAttribute( "questionfontsize", String.valueOf( newval ) );	
	}

	/** Gets the font size that this card container uses for review answers.
	 */
	public float getAnswerFontSize( ) {
		return Float.parseFloat( xml.getAttribute( "answerfontsize",
				String.valueOf(FontController.getUnicodeFont().getSize()) ) );
	}

	/** Sets the font size that this card container uses for review answers.
	 */
	public void setAnswerFontSize( int newval ) {
		xml.setAttribute( "answerfontsize", String.valueOf( newval ));		
	}
	
	/** Returns the number of instances in this box of the card with the
	  * given id.
	 */
	public int instanceCount( int cardid ) {
		int c = 0;
		Iterator i = getAllCards().iterator();
		Card curcard;
		while (i.hasNext()) {
			curcard = new Card( (IXMLElement)i.next() );
			if ( curcard.getID() == cardid ) c++;
		}
		return c;
	}
	
	/** Gets a vector containing references to every "card" xml tag in the box.
	 */
	public Vector getAllCards() {
		Vector rval = new Vector();
		for ( int c = 0; c <= (getPartitionCount()*2)+3; c++ ) {
			rval.addAll( getCardsInCompartment( c ) );
		}
		return rval;
	}
	
	/** Returns the compartment number of the first compartment containing a
	  * card with the given id, or -1 if no such compartment was found.
	 */
	public int findFirstCompartmentContainingCardID( int cardID ) {
		IXMLElement compartment;
		Vector ccards;
		Card ccard;
		for ( int c = 0; c < ((getPartitionCount()*2)+3); c++ ) {
			compartment = getCompartment( c );
			ccards = compartment.getChildrenNamed( "card" );
			for ( int b = 0; b < ccards.size(); b++ ) {
				ccard = new Card( (IXMLElement)ccards.get(b) );
				if ( ccard.getID() == cardID ) return c;
			}
		}
		return -1;
	}
	
	/** Gets ids and compartment numbers of all cards containing the given 
	  * text string in the question or answer.
	  * @param searchstring the String to search for in the cards.
	  * @param cset card set whose metacards' text we should search through.
	  * @return A vector of X arrays each containing two ints: the first is the
	  * compartment number of matching card X.  The second is the card ID of
	  * matching card X.
	  */
	public Vector findCardsContainingString(	String searchstring,
												CardSet cset ) { 
		int[][] allcardids = new int[2][ getCardCount() ];
		Vector curcompartmentcards;
		Card curcard;
		int j = 0;
		// Assemble an array of all cards ids in the box.
		for( int c = 0; c <= (getPartitionCount()*2)+3; c++ ) {
			curcompartmentcards = getCompartment( c ).getChildrenNamed("card");
			Iterator i = curcompartmentcards.iterator();
			while (i.hasNext()) {
				curcard = new Card( (IXMLElement)i.next() );
				allcardids[0][j] = c;
				allcardids[1][j] = curcard.getID();
				j++;
			}
		}
		
		// Search each card for the specified target string.
		Vector rval = new Vector();
		MetaCard mc;
		for (int c = 0; c < allcardids[0].length; c++ ) {
			mc = cset.getMetaCard( allcardids[1][c] );
			String text = mc.getAnswer() + mc.getQuestion();
			if (text.toUpperCase().indexOf(searchstring.toUpperCase()) != -1) {
				// Found the string.  Add this card to the return vector.
				int addme[] = new int[2];
				addme[0] = allcardids[0][c];
				addme[1] = allcardids[1][c];
				rval.add( addme );
			}
		}

		return rval;
	}
	
	/** Gets the number of partitions in this box.
	  * @return the number of partitions in this box.
	 */
	public int getPartitionCount( ) {
		return getPartitions( ).size( );
	}
	
	/** Returns the compartment number correpsonding to the given partition
	  * number.
	 */
	public int partitionNumberToCompartmentNumber( int partitionNumber ) {
		if ( partitionNumber < -2 || partitionNumber > getPartitionCount() )
			return -2;
		if ( partitionNumber == -2 ) return -2;
		if ( partitionNumber == -1 ) return 1; // prebox pile
		if ( partitionNumber == getPartitionCount( ) ) 
			return ( getPartitionCount() * 2 ) + 3; //post-box review pile
		return ( partitionNumber * 2 ) + 3;
	}
	
	/** Returns the partition number correpsonding to the given compartment
	  * number.
	 */
	public int compartmentNumberToPartitionNumber( int compartmentNumber ) {
		compartmentNumber -= 2;
		if ( (compartmentNumber%2) == 0 ) {
			return compartmentNumber/2;
		} else {
			return (compartmentNumber-1)/2;
		}			
	}
	
	/** Gets the total card capacity of the box's partitions.
	 */
	public int getCapacity( ) {
		int output = 0;
		int cMax = getPartitionCount( );
		for ( int c = 0; c < cMax; c++  ) {
			output += getPartitionCapacity( c );
		}
		return output;
	}
	
	/** Gets the total number of cards in the box (with pre/post box piles).
	 */
	public int getCardCount( ) {
		int output = 0;
		// the partitions, their waiting piles, and post and prebox and waiting:
		int cMax = ( getPartitionCount( ) * 2 ) + 4;
		for ( int c = 0; c < cMax; c++  ) {
			output += getCompartmentCardCount( c );
		}
		return output;
	}
	
	/** Gets the number of cards in a compartment.
	  * Returns 0 if an invalid compartment number is supplied.
	 */
	public int getCompartmentCardCount( int compartmentNumber ) {	
		if (	(compartmentNumber < 0) 
				|| (compartmentNumber > (getPartitionCount()*2)+3)) {
			return 0;
		}
		IXMLElement compartment = getCompartment( compartmentNumber );
		return ((Vector) compartment.getChildrenNamed( "card" ) ).size( );		
	}
	
	/** Gets a vector containing the IXMLElements reperesenting the cards in
	  * the given compartment. 
	 */
	public Vector getCardsInCompartment( int compartmentNumber ) {
		IXMLElement compartment = getCompartment( compartmentNumber );
		return ((Vector) compartment.getChildrenNamed( "card" ) );
	}

	/** Gets the number of cards in a partition.
	  * Returns 0 if an invalid partition number is supplied.
	 */
	public int getPartitionCardCount( int partitionNumber ) {		
		if (	(partitionNumber < -1) 
				|| (partitionNumber > getPartitionCount())) {
			return 0;
		}
		IXMLElement compartment = getPartition( partitionNumber );
		return ((Vector) compartment.getChildrenNamed( "card" ) ).size( );		
	}
		
	/** Gets the capacity of a compartment.
	 */
	public int getCompartmentCapacity( int compartmentNumber ) {
		IXMLElement compartment = getCompartment( compartmentNumber );
		return compartment.getAttribute( "capacity", 0 );				
	}
	
	/** Gets the capacity of a partition.
	 */
	public int getPartitionCapacity( int partitionNumber ) {
		int cNumber = partitionNumberToCompartmentNumber(partitionNumber);
		IXMLElement compartment = getCompartment( cNumber );
		return compartment.getAttribute( "capacity", 0 );
	}
	
	// Returns the xml tag of a compartment, based on its partition number.
	private IXMLElement getPartition( int partitionNumber ) {
		int cNumber = partitionNumberToCompartmentNumber( partitionNumber );
		return getCompartment( cNumber );
	}
	
	// Returns the xml tag of a compartment, based on its compartment number.
	private IXMLElement getCompartment( int compartmentNumber ) {

		if ( compartmentNumber == 0 ) {
			IXMLElement preboxxml = xml.getFirstChildNamed( "prebox" );
			if ( preboxxml == null )
				{ throw new RuntimeException("Bad box data."); }
			return getOrMakeWaitingTag( preboxxml );
		} 

		if ( compartmentNumber == 1 ) {
			IXMLElement preboxxml = xml.getFirstChildNamed( "prebox" );
			if ( preboxxml == null )
				{ throw new RuntimeException("Bad box data."); }
			return preboxxml;
		}
		
		int pCount = this.getPartitionCount( );
		int postboxNumber = ( pCount * 2 ) + 2;
		
		if ( compartmentNumber == postboxNumber ) {
			IXMLElement postboxxml = xml.getFirstChildNamed( "postbox" );
			if ( postboxxml == null )
				{ throw new RuntimeException("Bad box data."); }
			return postboxxml;
		}
		
		if ( compartmentNumber == postboxNumber+1 ) {
			IXMLElement postboxxml 
				= xml.getFirstChildNamed("postboxreviewpile");
			if ( postboxxml == null )
				{ throw new RuntimeException("Bad box data."); }
			return postboxxml;
		}
		
		// We are being asked for one of the inner compartments.

		// Note: this code assumes the partitions are stored in the box tag in
		// order from 0 - N.
		try {
			Vector parts = getPartitions( );
			compartmentNumber -= 2;
			if ( (compartmentNumber%2) == 0 ) {
				// We were asked for a partition waiting pile.
				IXMLElement p = (IXMLElement) parts.get( compartmentNumber/2 );
				return getOrMakeWaitingTag( p );
			} else {
				// We were asked for a partition.
				return (IXMLElement) parts.get( (compartmentNumber-1)/2 );
			}			
		} catch ( ArrayIndexOutOfBoundsException aioobe ) {
			return null;
		}

	} // end getCompartment
	
	/** Returns either the existing "waiting" subtag of the given parent tag, 
	  * or creates a new such subtag and returns it if none already exists.
	 */
	private IXMLElement getOrMakeWaitingTag( IXMLElement parent ) {
		IXMLElement output = parent.getFirstChildNamed( "waiting" );
		if ( output == null ) { 
			output = new XMLElement( "waiting" );
			parent.addChild( output );		
		}
		return output;
	}
	
	/** Shuffles the cards in the given compartment with a randomized insertion
	  * sort.
	  * @param compartmentNumber the compartment to shuffle
	 */
	public void shuffleCompartment( int compartmentNumber ) {
		IXMLElement compartment = getCompartment( compartmentNumber );
		Vector cards = compartment.getChildrenNamed( "card" );
		int ccount = cards.size( );
		
		while ( !cards.isEmpty( ) ) {
			int c = random.nextInt( cards.size( ) );
			IXMLElement moveme = (IXMLElement) cards.remove( c );
			// Remove a random card and stick it back on the end.
			compartment.removeChild( moveme );
			compartment.addChild( moveme );
		}
	}
	
	/** Adds the specified card to the pre-box pile.
	  * @param newCard the card to be added
	 */
	public void add( CommonCard newCard ) {
		if ( newCard instanceof StreamCard ) {
			newCard = ((StreamCard)newCard).toCard();
		}
		add((Card)newCard);
	}

	/** Adds the specified card to the pre-box pile.
	  * @param newCard the card to be added
	 */
	public void add( Card newCard ) {
		this.getCompartment( 1 ).addChild( newCard.toXML( ) );
	}

	/** Adds the specified card to the back of the specified compartment.
	  * @param newCard the card to be added
	 */
	public void add( Card newCard, int compartmentNumber ) {
		this.getCompartment( compartmentNumber ).addChild( newCard.toXML( ) );
	}
	
	/** Removes the specified card from the specified compartment.
	  * @param cardID ID of card to remove.
	  * @param compartmentNumber compartment number to remove card from
	  * @return true if the card was found and removed, false otherwise.
	 */
	public boolean remove( int cardID, int compartmentNumber ) {
		IXMLElement c = getCardByID( cardID, compartmentNumber );
		if ( c == null ) return false;
		IXMLElement compartment = getCompartment( compartmentNumber );
		compartment.removeChild( c );
		return true;
	}

	/** Removes the specified card if found.
	  * @param cardID ID of card to remove.
	  * @return true if the card was found and removed, false otherwise.
	 */
	public boolean remove( int cardID ) {
		int compartment = findFirstCompartmentContainingCardID( cardID );
		if ( compartment == -1 ) return false;
		return remove( cardID, compartment );
	}
	
	private IXMLElement getCardByID( int cardID, int compartmentNumber ) {
		if ( cardID < 0 ) return null;
		IXMLElement c = getCompartment( compartmentNumber );
		if ( c == null ) return null;
		Vector cards = c.getChildrenNamed( "card" );
		Iterator i = cards.iterator( );
		while ( i.hasNext() ) {
			IXMLElement x = (IXMLElement) i.next( );
			if ( x.getAttribute("id", "-1").compareToIgnoreCase( 
					String.valueOf(cardID) ) == 0 ) return x;
		}
		return null;
	}
	
	/** Moves a card from the specified compartment to the specified
	  * compartment.
	  * @param cardID ID of card to be moved
	  * @param from number of compartment to move the card from
	  * @param to number of compartment to move the card to
	  * @return true if the card was found and moved, false otherwise
	 */
	public boolean move( int cardID, int from, int to ) {
		IXMLElement c = getCardByID( cardID, from );
		if ( c == null ) return false;
		IXMLElement f = getCompartment( from );
		IXMLElement t = getCompartment( to );
		if ( t == null ) return false;
		f.removeChild( c );
		t.addChild( c );
		return true;		
	}

	public void beginReview( ) {
		beginReview( 0, true );
	}

	public void beginReview( int partitionNumber ) {
			beginReview( partitionNumber, true );
	}

	/** Begins a new review session at the specified partition number.
	  * 
	  * @param partitionNumber, -1 for the pre-box pile, (partition count) - 1
	  * for the post-box long term review pile.
	  * @param endreviewfirst should the current review session (if there is
	  * one) be ended before beginning the new one? 
	 */
	public void beginReview( int partitionNumber, boolean endreviewfirst ) {
		
		if ( endreviewfirst ) {

			endReview( false );
			
			// reset score
			setRight( 0 );
			setWrong( 0 );
		}
		
		// If we were not requested to end the review, and current compartment
		// is not -2, a review was already underway.  We should rebegin it.
		if ( getCurrentPartitionNumber() != -2 ) {
			return;
		}
		
		// Find the first partition containing cards after partitionNumber,
		// inclusive:
		int cMax = getPartitionCount();
		int c = partitionNumber;
		while ( c < cMax ) {
			if ( getPartitionCardCount( c ) > 0 ) break;
			c++;
		}
		
		if ( c == cMax ) {
			// No partitions have any cards to review. 
			c = -2; 
		} 

		setCurrentPartition( c );
	}

	/** Begins a new long term review session of all the post-box cards.
	 */
	public void beginLongTermReview( ) {
		beginLongTermReview( 
			getCompartmentCardCount( 
				getPostBoxCompartmentNumber( ) ) );
	}

	/** Begins a new long term review session of the specified number of.
	  * 
	  * @param ccount number of post-box cards to review (at max).
	 */
	public void beginLongTermReview( int ccount ) {

		endReview( true );
		
		if ( ccount > getPostBoxCompartmentNumber( ) )
			ccount = getPostBoxCompartmentNumber( );
			
		// Move a stack of post-box cards to post-box review pile.
		IXMLElement csrc = getCompartment( getPostBoxCompartmentNumber( ) );
		IXMLElement cdest = getCompartment(getPostBoxCompartmentNumber()+1);
		Vector srccards = csrc.getChildrenNamed( "card" );
		Iterator i = srccards.iterator( );
		IXMLElement curcard;
		while ( i.hasNext( ) ) {
			curcard = (IXMLElement)i.next();
			csrc.removeChild( curcard );
			cdest.addChild( curcard );
		}

		// reset score
		setRight( 0 );
		setWrong( 0 );
		
		if ( ccount == 0 ) {
			// No post-box cards to review. 
			setCurrentPartition( -2 );
			return; 
		} 
		
		setCurrentPartition( getPartitionCount() );
			
	}
	
	/** Ends any review currently in session, moving all waiting cards back
	  * into their compartments and shuffling the compartments.
	 */
	public void endReview( ) { endReview( true ); }

	/** Ends any review currently in session, moving all waiting cards back
	  * into their compartments and shuffling the compartments if specified.
	 */
	public void endReview( boolean reshuffle ) {
		xml.setAttribute( "currentpartition", "-2" ); // -2 = not reviewing 
		mergeInWaiting( reshuffle );
		IXMLElement statstag = xml.getFirstChildNamed( "reviewstatistics" );
		if ( statstag == null ) {
			statstag = new XMLElement( "reviewstatistics" );
			xml.addChild( statstag );
		}
		
		// record review statistics
		IXMLElement newtag = new XMLElement( "reviewsession" );
		newtag.setAttribute( "date", new Card().getDateString( ) );
		newtag.setAttribute( "right", String.valueOf( getRight( ) ) );
		newtag.setAttribute( "wrong", String.valueOf( getWrong( ) ) );
		statstag.addChild( newtag );
		orderedreviewsessions.add( new ReviewSessionReference( newtag ) );
		Collections.sort( orderedreviewsessions );
		
	}
	
	/** Sets the current review partition.
	  * @param newval the new value for the review partition.
	 */
	public void setCurrentPartition( int newval ) {
		xml.setAttribute( "currentpartition", String.valueOf( newval ) );
	}
	
	/** Gets the number of cards right so far this review session.
	 * @return the number of cards right so far this review session.
	 */
	public int getRight() {
		return xml.getAttribute( "right", 0 );
	}

	/** Sets the number of cards right so far this review session.
	  * @param newval the new number of cards right.
	 */
	public void setRight( int newval ) {
		xml.setAttribute( "right", String.valueOf( newval ) );
	}
	
	/** Gets the number of cards wrong so far this review session.
	 * @return the number of cards wrong so far this review session.
	 */
	public int getWrong() {
		return xml.getAttribute( "wrong", 0 );
	}

	/** Sets the number of cards right so far this review session.
	  * @param newval the new number of cards right.
	 */
	public void setWrong( int newval ) {
		xml.setAttribute( "wrong", String.valueOf( newval ) );
	}

	/** Gets the number of cards reviewed so far this review session.
	 * @return the number of cards reviewed so far this review session.
	 */
	public int getReviewed() {
		return (getRight( ) + getWrong( ));
	}

	/** Gets a string of the form "[right] / [total] ([percent right] %)" 
	  * describing the status of the current review session.
 	  * @return the status string.
	 */
	public String getReviewStatusString() {
		Float fval1 = Float.valueOf(getRight());
		Float fval2 = Float.valueOf(getReviewed());
		int percentCorrect = new Float 
			((fval1.floatValue() / fval2.floatValue())*100.0).intValue();
		String s =
			getRight() + " / " + getReviewed() + " (" +  percentCorrect + " %)";
		return s;
	}
	
	/** Gets a list of all review sessions recorded for this box.
	  * @return a list of all review sessions recorded for this box. If no
	  * review sessions are recorded, an empty list is returned.
	 */
	public List getReviewSessions( ) {
		List output = new Vector( );
		IXMLElement stats = xml.getFirstChildNamed( "reviewstatistics" );
		if ( stats == null ) { return output; }
		output = stats.getChildrenNamed( "reviewsession" );
		return output;
	}
	
	/** Gets the most recent review session recorded for this box.
	  * @return the most recent review session recorded for this box, null if
	  * no review sessions are recorded.
	 */
	public ReviewSession getLastReviewSession( ) {
		if ( orderedreviewsessions.size() < 1 ) { return null; }
		ReviewSessionReference ref = 
			(ReviewSessionReference)orderedreviewsessions.get( 
				orderedreviewsessions.size() - 1 );
		IXMLElement session = ref.xml; 
		if ( session == null ) { return null; }
		String sDate = session.getAttribute( "date", "No Date" );
		String sRight = session.getAttribute( "right", "0" );
		String sWrong = session.getAttribute( "wrong", "0" );
		return new ReviewSession( sDate, sRight, sWrong );
	}
	
	/** Moves all cards from the specified source compartment to the specified 
	  * destination compartment.
	  * @param from number of the source compartment
	  * @param to number of the destination compartment
	  * @return the number of cards moved
	 */
	public int moveAll( int from, int to ) {
		IXMLElement f = getCompartment( from );
		IXMLElement t = getCompartment( to ); 
		if ( f == null || t == null ) return 0;
		Vector fcards = f.getChildrenNamed( "card" );
		Iterator i = fcards.iterator( );
		int num = 0;
		while ( i.hasNext( ) ) {
			IXMLElement c = (IXMLElement) i.next( );
			f.removeChild( c );
			t.addChild( c );
			num++;
		}
		return num;
	}
	
	// Forces all waiting cards into their compartment.
	private void mergeInWaiting( boolean shuffle ) {
		// pre-box waiting pile -> pre-box
		moveAll( 0, 1 );
		if ( shuffle ) shuffleCompartment( 1 );
		int pcount = getPartitionCount( );
		for ( int p = 0; p < pcount; p++ ) {
			moveAll( ((p*2)+2), ((p*2)+3) );
			if ( shuffle ) this.shuffleCompartment( ((p*2)+3) );
		}
		// post-box review waiting pile -> post-box
		moveAll( ((pcount*2)+3), ((pcount*2)+2) );
	}
	
	/** Gets the partition number currently under review, or -2 if no partition
	  * is currently under review.
	  * @return The partition number currently under review, or -2 if there is
	  *  no review in progress, or -1 if the prebox pile is under review, or
	  *  the number of partitions in the box if the postbox review pile is
	  *  under review.
	 */
	public int getCurrentPartitionNumber( ) {
		return Integer.parseInt(xml.getAttribute( "currentpartition", "-2" ));
	}
	
	private IXMLElement getCurrentPartition( ) {
		return getPartition( getCurrentPartitionNumber( ) );
	}
	
	/** Gets the ID of the card under review (the frontmost card in the current
	  * partition under review).
	  * @return the ID of the card under review, or -1 if there is no review
	  *  in session.
	 */
	public int getCurrentCardID( ) {
		IXMLElement c = getCurrentCard( );
		if ( c == null ) return -1; 
		return c.getAttribute( "id", -1 );
	}
	
	private IXMLElement getCurrentCard( ) {
		IXMLElement p = getCurrentPartition( );
		if ( p == null ) return null;
		else {Vector v = p.getChildrenNamed( "card" );
			if ( v.size() == 0 ) return null;
			return (IXMLElement)v.get( v.size() - 1 );
		}
	}
		
	// Gets the number of the compartment under review, or -2 if there is none.
	public int getCurrentCompartmentNumber( ) {
		return partitionNumberToCompartmentNumber(getCurrentPartitionNumber());
	}
	
	// Gets the number of the target compartment for right cards.
	private int getNextCompartmentNumber( ) {		
		if ( getCurrentPartitionNumber() == getPartitionCount() ) {
			// special case if this is the post-box review pile.
			return getPostBoxCompartmentNumber( );
		}		
		return getCurrentCompartmentNumber( ) + 1;
	}
	
	// Gets the number of the postbox review compartment.
	private int getPostBoxCompartmentNumber() {
		return ( getPartitionCount() * 2 ) + 2;
	}
	
	/** Moves the card currently under review up to the next partition's
	  * waiting area, advancing the partition under review if all cards in the
	  * current one have been exhausted.
	 */
	public void right( ) {

		IXMLElement c = getCurrentCard( );
		int cardID = c.getAttribute( "id", -1 );
		if ( cardID == -1 ) return;

		String undoLastCardReviewDate = 
			c.getAttribute( "lastcardreviewdate", "..." );
		int undoSetCompartmentBackTo = getCurrentCompartmentNumber( );

		// update card stats
		c.setAttribute( "lastreviewdate", new Card().getDateString( ) );
		int streak = c.getAttribute( "streak", 0 );
		int undostreak = streak;
		if ( streak < 1 ) {
			streak = 1;
		} else {
			streak++;
		}
		c.setAttribute( "streak", String.valueOf( streak ) );
		c.setAttribute( "lifetimerights", 
				String.valueOf( c.getAttribute( "lifetimerights", 0 ) + 1 ) );

		// move the card
		int from = getCurrentCompartmentNumber( );
		int to = getNextCompartmentNumber( );
		move( cardID, from, to );
		
		// increment the score
		setRight( getRight( ) + 1 );
						
		// update box state
		advanceCompartmentIfNecessary();

		addUndoData(	cardID,
						undoLastCardReviewDate,
						"undo error",
						to,
						from,
						undoSetCompartmentBackTo,
						(getCurrentCompartmentNumber( ) == -2),
						undostreak,
						true );
	}
	
	/** Moves the card currently under review down to the prebox pile's waiting
	  * area, advancing the partition under review if all cards in the current
	  * partition have been exhausted.
	 */
	public void wrong( ) {
		
		IXMLElement c = getCurrentCard( );
		int cardID = c.getAttribute( "id", -1 );
		if ( cardID == -1 ) return;
		
		String undoLastCardReviewDate = 
			c.getAttribute( "lastcardreviewdate", "..." );
		int undoSetCompartmentBackTo = getCurrentCompartmentNumber( );
		
		String undoLastWrongDate =
			c.getAttribute( "lastwrongdate", "none" );
		
		// update card stats
		String ds = new Card().getDateString( );
		c.setAttribute( "lastreviewdate", ds );
		c.setAttribute( "lastwrongdate", ds );
		int streak = c.getAttribute( "streak", 0 );
		int undostreak = streak;
		if ( streak > 0 ) {
			streak = -1;
		} else {
			streak--;
		}
		c.setAttribute( "streak", String.valueOf( streak ) );
		c.setAttribute( "lifetimewrongs", 
				String.valueOf( c.getAttribute( "lifetimewrongs", 0 ) + 1 ) );

		// move the card
		int from = getCurrentCompartmentNumber( );
		move( cardID, from, 0 );
		
		// increment the score
		setWrong( getWrong( ) + 1 );
		
		// update box state
		advanceCompartmentIfNecessary( );

		addUndoData(	cardID,
						undoLastCardReviewDate,
						undoLastWrongDate,	
						0,
						from,
						undoSetCompartmentBackTo,
						(getCurrentCompartmentNumber( ) == -2),
						undostreak,
						false );
	}
	
	/** Gets the next undo command in the queue, or the single undo command
	 * if there is no queue.
	 */
	private UndoData getNextUndo( ) {
		UndoData output = undodata;
		undodata = null;
		return output;
	}
	
	/** Uses the specified data for undo requests. */
	private void addUndoData(	int cardid,
								String lastcardreviewdate,
								String lastwrongdate,
								int fromcompartment,
								int tocompartment,
								int setcompartmentbackto,
								boolean removelastreviewrecord, 
								int streak,
								boolean itwasright ) {

		UndoData ud = new UndoData( );
		ud.cardid = cardid;
		ud.lastcardreviewdate = lastcardreviewdate;
		ud.lastwrongdate = lastwrongdate;
		ud.fromcompartment = fromcompartment;
		ud.tocompartment = tocompartment;
		ud.setcompartmentbackto = setcompartmentbackto;
		ud.removelastreviewrecord = removelastreviewrecord;
		ud.streak = streak;
		ud.itwasright = itwasright;

		undodata = ud;
	}
	
	// Moves the current compartment under review up if necessary.
	private void advanceCompartmentIfNecessary( ) {
		IXMLElement ccomp = getCurrentPartition();
		int pnumber = getCurrentPartitionNumber();
		
		// If the next compartment is the last one
		// or the number of cards waiting for the next partition
		// are less than the size of the next partition
		// and the current partition is not empty, 
		// then do not advance.
		if (((getNextCompartmentNumber() == getPostBoxCompartmentNumber())
			|| (getCompartmentCapacity(getNextCompartmentNumber() + 1) >
				getCompartmentCardCount(getNextCompartmentNumber())))
			&& ( getCurrentCard() != null )) {
			return;				
		}
		
		// else we need to advance.
		if ( pnumber == getPartitionCount( ) ) {
			// finished the post-box review pile... review is done.
			xml.setAttribute( "currentpartition", "-2" );
			endReview();
			return;
		}
		if ( pnumber == (getPartitionCount( )-1) ) {
			// finished the last partition... review is done.
			xml.setAttribute( "currentpartition", "-2" );
			endReview();
			return;
		}
		// move up one partition, attempt to add its waiting cards, and continue
		// review there if the waiting cards will not fit.  If the cards do fit,
		// end the review.
		pnumber++;
		if ( tryToMergeInWaiting( pnumber ) ) {
			xml.setAttribute( "currentpartition", "-2" );
			endReview();
			return;			
		} else {
			// The waiting cards won't fit.
			// If this partition is empty, force them in and end the review
			// session.  If this partition is not empty, review it.
			xml.setAttribute( "currentpartition", String.valueOf( pnumber ) );
			if ( getCurrentCard( ) == null ) {
				xml.setAttribute( "currentpartition", "-2" );
				endReview();
				return;			
			}
		}
	}

	// Returns true if the waiting pile of the given partition fit into the
	// partition and was merged in, or false if it did not and was not.
	private boolean tryToMergeInWaiting( int partitionNumber ) {
		int compartmentNumber = 
			partitionNumberToCompartmentNumber( partitionNumber );
		int from = compartmentNumber-1;
		int to = compartmentNumber;
		
		if ( (getCompartmentCapacity( to ) - getCompartmentCardCount( to ) )
			 >= getCompartmentCardCount( from ) ) {
			// There is room.
			moveAll( from, to );
			return true;
		} else {
			// There is not room.
			return false;
		}
	}

	/** Returns true if the last action on this card container can be undone.
	 */
	public boolean canUndo() {
		return (undodata != null);
	}
	
	/** Undoes the effects of the last Right( ) or Wrong( ) call.
	  * Returns true if the undo was successful.
	 */
	public boolean Undo( ) {
		
		UndoData ud = getNextUndo();		
		if ( ud == null ) return false;
		IXMLElement c = getCardByID( ud.cardid, ud.fromcompartment );		
		if ( c == null ) return false;
				
		// update card stats
		c.setAttribute( "lastreviewdate", ud.lastcardreviewdate );
		c.setAttribute( "streak", String.valueOf( ud.streak ) );
		
		if ( ud.itwasright ) {
			setRight( getRight( ) - 1 );
			c.setAttribute( "lifetimerights", 
					String.valueOf(c.getAttribute("lifetimerights", 0) - 1 ));
		} else {
			setWrong( getWrong( ) - 1 );
			c.setAttribute( "lifetimewrongs", 
					String.valueOf(c.getAttribute("lifetimewrongs", 0) - 1 ));
			c.setAttribute( "lastwrongdate", ud.lastwrongdate );
		}
		
		// move the card
		move( ud.cardid, ud.fromcompartment, ud.tocompartment );
		
		// undo review session entry if necessary
		if ( ud.removelastreviewrecord ) {
			IXMLElement rrecs = xml.getFirstChildNamed( "reviewstatistics" );
			Vector sessions = rrecs.getChildrenNamed( "reviewstatistics" );
			rrecs.removeChildAtIndex( sessions.size( ) - 1 );
		}

		// update box state
		setCurrentPartition( 
			compartmentNumberToPartitionNumber( ud.setcompartmentbackto ) );
		
			
		return true;
	}

	/** Holds data about what tasks one "undo" action should perform.
	 */
	private class UndoData {
		public int cardid;
		public String lastcardreviewdate;
		public String lastwrongdate;
		public int fromcompartment;
		public int tocompartment;
		public int setcompartmentbackto;
		// if undo involves restarting a review session:
		public boolean removelastreviewrecord;
		public int streak;
		public boolean itwasright;
	}
	
	/** For keeping an chronologically ordered list of review sessions.
	 */
	private static class ReviewSessionReference implements Comparable {

		public IXMLElement xml;
		private String ordernumber;
		
		public ReviewSessionReference( IXMLElement newxml ) {
			xml = newxml;
			ordernumber = dateStringToOrderNumber( 
				xml.getAttribute( "date", "none" ) );
		}
		
		public static String dateStringToOrderNumber( String s ) {
			String[] datecomponents = s.split( " " );
			if ( datecomponents.length != 3 ) return "0";
			String[] component2 = datecomponents[1].split( ":" );
			String ordernumber = datecomponents[0];
			ordernumber += datecomponents[2]; // AM/PM
			ordernumber += component2[0] + component2[1];
			return ordernumber;
		}

		// > 0 if this refers to a newer review session than other. 
		public int compareTo( Object other ) {
			return ordernumber.compareTo( 
				((ReviewSessionReference)other).ordernumber);
		}
	}

	/** Returns a new panel of the class that should be used to display this
	  * card container type (e.g. BoxDisplayPanel or StreamDisplayPanel.)
	 */
	public CardContainerDisplayPanel getNewDisplay() {
		return new BoxDisplayPanel();
	}

	/** Returns a new card browsing table panel of the appropriate class for
	  * this card container, using the given data to fill it in first.
	 */
	public CommonCardTablePanel getNewTablePanel( 
			Vector clist, CardSet cs, AnkiStreamFrame asf ) {
		return new CardTablePanel( clist, cs, asf );
	}
	
} // end class Box
