package ankistream;

import java.awt.Dimension;
import java.awt.Font;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.event.MouseListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent; 
import java.util.Vector;

import javax.swing.JPanel;
import javax.swing.BorderFactory;
import javax.swing.JOptionPane;

import net.n3.nanoxml.IXMLElement;

/** A panel displaying a anki box and its cards visually.  Shown on the
  * right hand side of the main frame of the Anki Box application.
  * <br> 
  * <br> Part of the Anki Stream project, a tool for long-term flash card
  * review.  Anki Stream is distributed under the MIT/X11 free software 
  * license, a copy of which should have come with this distribution.
  * @author Bishop Wilkins
  * @version 0.1
*/ 
public class BoxDisplayPanel 
	extends JPanel implements CardContainerDisplayPanel {
	
	private Box box;
	private BufferedImage boximage; // Double buffer for drawing the box.
	private int selectedCompartment; // -1 if no compartment is selected.
	private AnkiStreamFrame ownerframe;
	
	/** Creates a new BoxDisplayPanel.
	 */
	public BoxDisplayPanel( ) {
		ownerframe = null;
		setPreferredSize( new Dimension( 320, 480 ) );
		setBorder( BorderFactory.createTitledBorder( "" ) );
		selectedCompartment = -1;
		addMouseListener(
			new MouseAdapter() {
				public void mouseClicked( MouseEvent e ) {
					 onClick( e.getPoint().x, e.getPoint().y );
				}
			}
		);
	}
	
	/** Begins invoking specified AnkiBoxFrame's onSelectionChanged()
	  * method whenever a new compartment is selected/deselected. 
	 */
	public void notifyMeOnSelection( AnkiStreamFrame newowner ) {
		ownerframe = newowner;
	}
	
	/** Paints this panel.
	 */
	public void paintComponent( Graphics g ) {

		super.paintComponent( g ); 

		if ( boximage != null  ) {
			g.drawImage( boximage, 0, 0, null );
		}
	}

	/** Redraws the graphical representation of the box.
	 */
	public void redraw( ) {
		
		// Resize the image if the panel size has changed.
		int width = getWidth( );
		int height = getHeight( );
		boximage = new BufferedImage(	width, height, 
										BufferedImage.TYPE_INT_RGB );
												
		// Prepare to draw to the double buffer.
		Graphics2D g = boximage.createGraphics( );
		g.setColor( getBackground( ) );
		g.fillRect( 0, 0, width, height );
		
		// Draw the box.
		Color boxforeground = ( new Color( 128, 64, 0 ) );
		Color boxbackground = ( new Color( 80, 40, 0 ) );
		Color cardcolor = ( Color.WHITE );
		Color highlightcolor = ( Color.RED );
		Color textcolor = ( Color.BLACK );
		
		int boxxpos = (int) (width * 0.42);
		int boxypos = (int) (height * 0.2);
		int boxheight = (int) (height * 0.6);
		int boxwidth = (int) (width * 0.4 );
		
		g.setColor( boxbackground );
		g.fillRect( boxxpos, boxypos, boxwidth, boxheight );
		g.setColor( boxforeground );
		g.drawRect( boxxpos, boxypos, boxwidth, boxheight );
		g.drawRect( boxxpos-1, boxypos-1, boxwidth+1, boxheight+1 );
		
		// Figure out the display width of one card in pixels.
		int boxcards = box.getCapacity( );
		int partitioncount = box.getPartitionCount( );
		double cardwidth = (double) boxheight / (double) boxcards;
		
		// If this is a review session, the selected compartment should be
		// the current compartment.
		if ( box.getCurrentPartitionNumber() != -2 ) {
			selectedCompartment = box.getCurrentCompartmentNumber( );
		}
		
		// Draw cards.
		g.setColor( cardcolor );
		int cardstart;
		int cardend = boxypos + boxheight - 1;
		for ( int c = 0; c < partitioncount; c++ ) {

			if ( selectedCompartment == 
					box.partitionNumberToCompartmentNumber( c ) ) {
				g.setColor( highlightcolor );
			} else {
				g.setColor( cardcolor );
			} 
			
			// Partition cards.
			int cardstodraw = box.getPartitionCardCount( c );
			if ( cardstodraw > 0 ) {
				if ( cardstodraw > box.getPartitionCapacity( c ) ) {
					cardstodraw = box.getPartitionCapacity( c );
				}
				cardstart = cardend - 
							(int) ((double) cardstodraw * cardwidth);
				if ( cardstart == cardend ) cardstart = cardend - 1;

				g.fillRect( boxxpos+1, cardstart, 
							boxwidth-2, (cardend-cardstart) );
			}
			
			// Partition waiting area cards.
			g.setColor( cardcolor );
			if ( selectedCompartment == (c*2)+2 ) {
				g.setColor( highlightcolor );
			}
			cardstodraw = box.getCompartmentCardCount( (c*2)+2 );
			if ( cardstodraw > 0 ) {
				if ( cardstodraw > box.getPartitionCapacity( c ) ) {
					cardstodraw = box.getPartitionCapacity( c );
				}
				cardstart = cardend - 
							(int) ((double) cardstodraw * cardwidth);
				if ( cardstart == cardend ) cardstart = cardend - 1;
				if ( cardstodraw == box.getPartitionCapacity( c ) ) {
					cardstart += 2;
				}

				g.fillRect( (int) ((double)width * 0.01), cardstart, 
							boxwidth-2, (cardend-cardstart) );
			}
			
			cardend -= 
				(int) ((double) box.getPartitionCapacity( c ) * cardwidth);
		}
		
		
		// Draw card count labels.
		g.setColor( textcolor );
		g.setFont( new Font( "MS UI Gothic", Font.PLAIN, 12 ) );
		int textxpos = boxxpos + boxwidth + 5;
		int waitingtextxpos = (int) ((double) width * 0.2);
		int textypos = boxypos + boxheight + 10;

		for ( int c = 0; c < partitioncount; c++ ) {
			int yinterval = 
				(int) ((double) box.getPartitionCapacity( c ) * cardwidth);
			yinterval /= 2;
			if ( yinterval < 8 ) textypos -= 8;
			else textypos -= yinterval;

			// Partition cards.
			String s = String.valueOf( box.getPartitionCardCount( c ) );
			s += " / ";
			s += String.valueOf( box.getPartitionCapacity( c ) );
			g.drawString( s, textxpos, textypos );
			
			// Waiting area for Partition
			if ( box.getCompartmentCardCount((c*2)+2) > 0 ) {
				s = String.valueOf( box.getCompartmentCardCount((c*2)+2) );
				g.drawString( s, waitingtextxpos, textypos );			
			}

			textypos -= yinterval;

		}
		
		//prebox waiting
		if ( box.getCompartmentCardCount( 0 ) > 0 ) {
			
			g.setColor( cardcolor );
			if ( selectedCompartment == 0 ) g.setColor( highlightcolor );
			
			int stackheight =
				(int) ((double) box.getCompartmentCardCount( 0 )*cardwidth);
			if ( stackheight < 1 ) stackheight = 1; 
			g.fillRect( (int) ((double)width * 0.01), boxypos+boxheight+10, 
						boxwidth-2, stackheight );
			
			String s = String.valueOf( box.getCompartmentCardCount( 0 ) );
			g.setColor( textcolor );
			g.drawString(	s, (int) ((double) width * 0.2), 
							boxypos + boxheight + 30 );
		}
		
		//prebox
		if ( box.getCompartmentCardCount( 1 ) > 0 ) {
			
			if ( selectedCompartment == 1 ) {
				g.setColor( highlightcolor );
			} else {
				g.setColor( cardcolor );
			} 
			
			int stackheight =
				(int) ((double) box.getCompartmentCardCount( 1 )*cardwidth);
			if ( stackheight < 1 ) stackheight = 1; 
			g.fillRect( boxxpos+1, boxypos+boxheight+10, 
						boxwidth-2, stackheight );
			
			String s = String.valueOf( box.getCompartmentCardCount( 1 ) );
			g.setColor( textcolor );
			g.drawString(	s, boxxpos + (boxwidth/2), 
							boxypos + boxheight + 30 );
		}

		//postbox
		int postboxcompartment = (partitioncount*2)+2;
		if ( box.getCompartmentCardCount( postboxcompartment ) > 0 ) {

			g.setColor( cardcolor );
			if (selectedCompartment == (box.getPartitionCount()*2)+2)
				g.setColor( highlightcolor );

			int stackbottom = boxypos-10;
			int stackheight =
				(int) ((double) box.getCompartmentCardCount( 
						postboxcompartment ) * cardwidth);
			if ( stackheight < 1 ) stackheight = 1; 
			int stacktop = stackbottom - stackheight;
			g.fillRect( boxxpos+1, stacktop, 
						boxwidth-2, stackheight );

			String s = 
				String.valueOf( 
					box.getCompartmentCardCount( postboxcompartment ) );
			g.setColor( textcolor );
			g.drawString(	s, boxxpos + (boxwidth/2), 
							boxypos - 45 );
		}

		//postboxreview
		if ( box.getCompartmentCardCount( postboxcompartment+1 ) > 0 ) {
			
			if ( selectedCompartment == (box.getPartitionCount()*2)+3 ) {
				g.setColor( highlightcolor );
			} else {
				g.setColor( cardcolor );
			} 

			int stackbottom = boxypos-10;
			int stackheight =
				(int) ((double) box.getCompartmentCardCount( 
						postboxcompartment+1 ) * cardwidth);
			if ( stackheight < 1 ) stackheight = 1; 
			int stacktop = stackbottom - stackheight;
			g.fillRect( (int) ((double)width * 0.01), stacktop, 
						boxwidth-2, stackheight );
			
			String s = 
				String.valueOf( 
					box.getCompartmentCardCount( postboxcompartment+1 ) );
			g.setColor( textcolor );
			g.drawString(	s, (int) ((double) width * 0.2), 
							boxypos - 45 );
		}
		
		// Draw each partition.
		int dividerpos = boxypos + boxheight;
		g.setColor( boxforeground );
		for ( int c = 0; c < (partitioncount-1); c++ ) {
			dividerpos -= 
				(int) ((double) box.getPartitionCapacity( c ) * cardwidth);
				g.fillRect( boxxpos, dividerpos-1, 
							boxwidth, 2 );
		}
				
		this.repaint( );		

	} // end redraw
	
	/** Called whenever the user clicks the mouse on the display panel.
	  * Tests whether the user clicked on a compartment and if so selects that
	  * component unless a review is underway.
	  * @param x x position the mouse was clicked at. Relative to this panel.
	  * @param y y position the mouse was clicked at. Relative to this panel.
	 */
	public void onClick( int x, int y ) {
		
		int cchoice = getCompartmentAt( x, y );

		// Don't select compartments with no cards in them.
		if ( box.getCompartmentCardCount( cchoice ) < 1 ) {
			cchoice = -1;
		}
		
		// If this is not a review session, select or deselect and highlight
		// according to the click.
		if ( box.getCurrentPartitionNumber() == -2) {
			if ( selectedCompartment == cchoice ) selectedCompartment = -1;
			else selectedCompartment = cchoice;
			redraw();
			if ( ownerframe != null ) {
				ownerframe.onSelectionChanged();
			}
		}
	}

	/** Returns the id of the selected card, or -1 if no card is selected.
	 */
	public int getSelectedCardID() {
		if (selectedCompartment == -1) {return -1;}
		Vector v =
			box.getCardsInCompartment(selectedCompartment);
		if ( v.size() > 0 ) {
			return (new Card((IXMLElement)v.firstElement())).getID();
		} else {
			return -1;
		}
	}
	
	/** Gets a vector of IXMLElements representing the cards around and 
	  * including the selected card to browse.
	 */
	public Vector getCardRange() {
		return box.getCardsInCompartment(selectedCompartment);
	}
	
	/** Gets the compartment at position (x, y) relative to this panel.
	  * or -1 if there was is no compartment there.
	 */
	private int getCompartmentAt( int x, int y ) {
		
		int rval = -1;

		int width = getWidth( );
		int height = getHeight( );

		int boxxpos = (int) (width * 0.42);
		int boxypos = (int) (height * 0.2);
		int boxheight = (int) (height * 0.6);
		int boxwidth = (int) (width * 0.4 );

		int boxcards = box.getCapacity( );
		int partitioncount = box.getPartitionCount( );
		double cardwidth = (double) boxheight / (double) boxcards;
		
		if ( x < (boxxpos - boxwidth - 6) ) {
			// to the left of any compartment
			rval = -1;

		} else if ( x < (boxxpos - 4) ) {
			// in the waiting cards column
			if ( y < (boxypos - 8) ) {
				// in the post-box review pile
				rval = ((box.getPartitionCount()*2)+3);
			} else if ( y < (boxypos + boxheight) ) {
				// in one of the partitions' waiting piles
				int dividertop = boxypos + boxheight;
				rval = -1;
				for ( int c = 0; c < partitioncount; c++ ) {
					dividertop -= 
						(int) ((double) box.getPartitionCapacity( c ) 
								* cardwidth);
					if ( y > dividertop ) {
						rval = ((c*2)+2);
						break;
					}
				}
			} else {
				// in the pre-box waiting pile
				rval = 0;
			}
			
		} else if ( x < (boxxpos + 2) ) {
			// between the waiting cards and the box partitions
			rval = -1;

		} else if ( x < (boxxpos + boxwidth - 2) ) {
			// in the box partitions column
			if ( y < (boxypos - 8) ) {
				// in the post-box pile
				rval = ((box.getPartitionCount()*2)+2);
			} else if ( y < (boxypos + boxheight) ) {
				// in one of the box partitions
				int dividertop = boxypos + boxheight;
				rval = -1;
				for ( int c = 0; c < partitioncount; c++ ) {
					dividertop -= 
						(int) ((double) box.getPartitionCapacity( c ) 
								* cardwidth);
					if ( y > dividertop ) {
						rval = box.partitionNumberToCompartmentNumber( c );
						break;
					}
				}

			} else {
				// in the pre-box pile
				rval = 1;
			}

		} else {
			// to the right of any compartment
			rval = -1;
		}
				
		return rval;
	}
	
	/** Returns the currently selected compartment.  Or -1 if none are selected.
	 */
	public int getSelectedCompartment( ) {
		return selectedCompartment;
	}
	
	/** Attempts to select the given compartment.
	  * @return true if the compartment was found, had cards, and was selected.
	  * false otherwise.
	 */
	public boolean selectCompartment( int compartment ) {
		if ( box.getCurrentCardID() != -1 ) return false;
		if (	compartment < 0 
				|| compartment > (box.getPartitionCount()*2)+3) return false;
		if ( box.getCompartmentCardCount( compartment ) < 1 ) return false;
		selectedCompartment = compartment;
		redraw();
		return true;
	}

	/** Gives this renderer an interface to access its card container.
	 */	
	public void setCardContainer( CardContainer cc ) {
		setBox((Box)cc);
	}
	
	/** Gives this renderer an interface to access its box.
	 */	
	public void setBox( Box newbox ) {		
		box = newbox;
	} // end setBox
	
} // end class BoxDisplayPanel
