package ankistream;

import java.util.Date;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;

/** <code>Card</code> represents one flash card in a box.  Cards have ID numbers
  * that correspond to the MetaCards which actually contain their display
  * information, and a set of statistics about review history.
  * <br> 
  * <br> Part of the Anki Stream project, a tool for long-term flash card
  * review.  Anki Stream is distributed under the MIT/X11 free software 
  * license, a copy of which should have come with this distribution.
  * @author Bishop Wilkins
  * @version 0.1
**/

public class Card implements CommonCard {
	
	private int id = 0;
	private String lastReviewDate;
	private String lastWrongDate;
	private int streak;
	private int lifetimeRights;
	private int lifetimeWrongs;
	private String creationDate;

	IXMLElement xml;

	public static final SimpleDateFormat dateform = 
		new SimpleDateFormat( "yyyy.MM.dd hh:mm aa" );

	/** Creates a new Card based on the specified XML "card" tag.
	  * @param iXML an xml "card" tag containing data about a card.
	 */
	public Card( IXMLElement iXML ) {
		xml = iXML;
		setID( xml.getAttribute( "id", 0 ) );
		setLastReviewDate( xml.getAttribute( "lastreviewdate", "unreviewed" ) );
		setLastWrongDate( xml.getAttribute( "lastwrongdate", "never missed" ) );
		setStreak( xml.getAttribute( "streak", 0 ) );
		setLifetimeRights( xml.getAttribute( "lifetimerights", 0 ) );
		setLifetimeWrongs( xml.getAttribute( "lifetimewrongs", 0 ) );
		setCreationDate( xml.getAttribute( "creationdate", getDateString( ) ) );
	}
	
	/** Creates a new uninitialized card. */
	public Card( ) {
		xml = new XMLElement( "card" );
		setLastReviewDate( "unreviewed" );
		setLastWrongDate( "never missed" );
		setStreak( 0 );
		setLifetimeRights( 0 );
		setLifetimeWrongs( 0 );
		setCreationDate( getDateString( ) );
	}
	
	/** Gets the xml element of this card.  Should only be called by the
	  * Anki Box framework when a new card is being created. */
	public IXMLElement toXML( ) { return xml; }
	
	/** Gets the ID of this card. */
	public int getID() { return id; }
	
	/** Gets the last review date of this card. */
	public String getLastReviewDate( ) { return lastReviewDate; }

	/** Gets the last date this card was missed in review. */
	public String getLastWrongDate( ) { return lastWrongDate; }

	/** Returns a positive number representing the number of times this card
	  * has been correctly identified in review, or a negative number
	  * representing the number of times it has been missed. */
	public int getStreak( ) { return streak; }

	/** Gets the number of times this card has been correctly identified. */
	public int getLifetimeRights( ) { return lifetimeRights; }

	/** Gets the number of times this card has been missed. */
	public int getLifetimeWrongs( ) { return lifetimeWrongs; }

	/** Gets the creation date of this card. */
	public String getCreationDate() { return creationDate; }
	
	
	/** Sets the ID of this card. */
	public void setID( int newID ) { 
		id = newID; 
		xml.setAttribute( "id", String.valueOf( newID ) );
	}

	/** Sets the last review date of this card. */
	public void setLastReviewDate( String val ) { 
		lastReviewDate = val; 
		xml.setAttribute( "lastreviewdate", lastReviewDate );
	}

	/** Sets the last date this card was missed in review. */
	public void setLastWrongDate( String val ) { 
		lastWrongDate = val; 
		xml.setAttribute( "lastwrongdate", lastWrongDate );
	}

	/** Sets the positive number representing the number of times this card
	  * has been correctly identified in review, or the negative number
	  * representing the number of times it has been missed. */
	public void setStreak( int val ) { 
		streak = val;
		xml.setAttribute( "streak", String.valueOf( streak ) );
	}

	/** Sets the number of times this card has been correctly identified. */
	public void setLifetimeRights( int val ) { 
		lifetimeRights = val;
		xml.setAttribute( "lifetimerights", String.valueOf( lifetimeRights ) );
	}

	/** Sets the number of times this card has been missed. */
	public void setLifetimeWrongs( int val ) { 
		lifetimeWrongs = val; 
		xml.setAttribute( "lifetimewrongs", String.valueOf( lifetimeWrongs ) );
	}

	/** Sets the creation date of this card. */
	public void setCreationDate( String val ) { 
		creationDate = val; 
		xml.setAttribute( "creationdate", creationDate );
	}
	
	/** Converts to a StreamCard. */
	public StreamCard toStreamCard() {
				
		StreamCard sc = new StreamCard( );
		sc.setID( getID() );
		sc.setCreationDate(convertDate(getCreationDate()));
		sc.setLastReviewDate(convertDate(getLastReviewDate()));
		
		// Estimate best next review time based on last wrong date or creation.
		Date nextreviewtime = new Date();
		long t1 = 0;
		long t2 = nextreviewtime.getTime();
		if ( lastWrongDate.equalsIgnoreCase( "never missed" ) ) {
			try {
				t1 = Card.dateform.parse(lastWrongDate).getTime();
			} catch ( ParseException pe ) { t1 = 0; }
			
		} else {
			try {
				t1 = Card.dateform.parse(getCreationDate()).getTime();
			} catch ( ParseException pe ) { t1 = 0;}
		}
		long span = t2 - t1;
		nextreviewtime = new Date(t2 + span);
		sc.setBestNextReviewDate(StreamCard.dateform.format(nextreviewtime));
		
		// Invent a history for this card.
		StringBuffer sb = new StringBuffer("");
		int rights = Math.min( getLifetimeRights(), 100 );
		int wrongs = Math.min( getLifetimeWrongs(), 100 );
		if ( getStreak() > 0 ) { 
			for ( int i = 0; i < Math.abs(getStreak()); i++ ) {
				sb.insert(0, '1');
			}
			rights -= getStreak();
			while ( (rights > 0) || (wrongs > 0) ) {
				if ( wrongs > 0 ) {sb.insert(0, '0'); wrongs--;}
				if ( rights > 0 ) {sb.insert(0, '1'); rights--;}
			}
		} else {
			for ( int i = 0; i < Math.abs(getStreak()); i++ ) {
				sb.insert(0, '0');
			}
			wrongs += getStreak();
			while ( (rights > 0) || (wrongs > 0) ) {
				if ( rights > 0 ) {sb.insert(0, '1'); rights--;}
				if ( wrongs > 0 ) {sb.insert(0, '0'); wrongs--;}
			}
		}
		sc.setLastHundredHistory(sb.toString());
		
		sc.setLastWrongDate(convertDate(getLastWrongDate()));
		sc.setLifetimeRights(getLifetimeRights());
		sc.setLifetimeWrongs(getLifetimeWrongs());
		
		// Make up an average forget period (= current time - last wrong time)
		String s = sc.getLastWrongDate();
		if ( s.equalsIgnoreCase( "never missed" ) ) {
			sc.setAverageForgetPeriod("never missed");
		} else {
			long lwt;
			try {
				lwt = (StreamCard.dateform.parse(s)).getTime();
			} catch (ParseException pe) {
				lwt = new Date().getTime();
			}
			long ct = new Date().getTime();
			if ( lwt > ct ) { lwt = ct; }
			long period = ct - lwt;
			sc.setAverageForgetPeriod(StreamCard.formatTimeSpan(period));
		}		
		
		return sc;
	}

	private String convertDate(String cardformat) {
		try {
			return StreamCard.dateform.format(Card.dateform.parse(cardformat));
		} catch ( ParseException pe ) {
			return cardformat;
		}
	}
	
	/** Gets the date in yyyy.mm.dd format. */
	public String getDateString() {
		return dateform.format( new Date( ) );
	}

} // end class Card
