package ankistream;

import java.util.Vector;
import net.n3.nanoxml.IXMLElement;

/** Since Anki Stream can open and work with different types of card collection
  * (streams and boxes for now), this is an interface that defines what
  * functions those collections must provide.
  * <br> 
  * <br> Part of the Anki Stream project, a tool for long-term flash card
  * review.  Anki Stream is distributed under the MIT/X11 free software 
  * license, a copy of which should have come with this distribution.
  * @author Bishop Wilkins
  * @version 0.1
**/

public interface CardContainer {
	
	public static final int DEFAULT_SECONDS_PER_CARD = 0;
	
	/** Gets the root of an XML tree containing the data for this stream.
	  * @return the root of an XML tree containing the data for this stream.
	 */
	public IXMLElement toXML( );
	
	/** Gets the name of this stream.
	  * @return the name of this stream
	 */
	public String getName( );
	
	/** Sets the name of this stream.
	 */
	public void setName( String newname );

	/** Gets the name of the data file containing the card set for this stream.
	 */
	public String getCardSetFile( );
	
	/** Sets the name of the data file containing the card set for this stream.
	 */
	public void setCardSetFile( String newfile );

	/** Gets the seconds per card that this card container uses for reviews.
	 */
	public int getSecondsPerCard( );

	/** Sets the seconds per card that this card container uses for reviews.
	 */
	public void setSecondsPerCard( int newval );

	/** Gets the font size that this card container uses for review questions.
	 */
	public float getQuestionFontSize( );

	/** Sets the font size that this card container uses for review questions.
	 */
	public void setQuestionFontSize( int newval );

	/** Gets the font size that this card container uses for review answers.
	 */
	public float getAnswerFontSize( );

	/** Sets the font size that this card container uses for review answers.
	 */
	public void setAnswerFontSize( int newval );

	/** Returns the number of instances in this stream of the card with the
	  * given id.
	 */
	public int instanceCount( int cardid );
		
	/** Gets ids of all cards containing the given text string in the question 
	  * or answer.
	  * @param searchstring the String to search for in the cards.
	  * @param cset card set whose metacards' text we should search through.
	  * @return A vector of ints representing the IDs of the cards containing
	  * the queriedstring.
	*/
	public Vector findCardsContainingString(	String searchstring,
												CardSet cset );
		
	/** Gets the number of cards in the stream.
	 */
	public int getCardCount( );
		
	/** Adds the specified card to the stream.
	  * @param newCard the card to be added
	 */
	public void add( CommonCard newCard );
	
	/** Removes the specified card from the stream.
	  * @param cardID ID of card to remove.
	  * @return true if the card was found and removed, false otherwise.
	 */
	public boolean remove( int cardID );
			
	/** Begins a new review session. */
	public void beginReview( );
	
	/** Ends any review currently in session.
	 */
	public void endReview( );
	
	/** Gets the number of cards right so far this review session.
	 * @return the number of cards right so far this review session.
	 */
	public int getRight();
	
	/** Sets the number of cards right so far this review session.
	  * @param newval the new number of cards right.
	 */
	public void setRight( int newval );
	
	/** Gets the number of cards wrong so far this review session.
	 * @return the number of cards wrong so far this review session.
	 */
	public int getWrong();
	
	/** Sets the number of cards right so far this review session.
	  * @param newval the new number of cards right.
	 */
	public void setWrong( int newval );
	
	/** Gets the number of cards reviewed so far this review session.
	 * @return the number of cards reviewed so far this review session.
	 */
	public int getReviewed();
	
	/** Gets a string of the form "[right] / [total] ([percent right] %)" 
	  * describing the status of the current review session.
 	  * @return the status string.
	 */
	public String getReviewStatusString();

	/** Gets the ID of the card under review (the front card in the stream).
	  * @return the ID of the card under review, or -1 if there is no review
	  *  in session.
	 */
	public int getCurrentCardID( );
	
	/** Called when the current card under review is declared right.
	 */
	public void right( );	
	/** Called when the current card under review is declared wrong.
	 */
	public void wrong( );
	
	/** Returns true if the last action on this card container can be undone.
	 */
	public boolean canUndo( );

	/** Undoes the effects of the last Right( ) or Wrong( ) call.
	  * Returns true if the undo was successful.
	 */
	public boolean Undo( );
	
	/** Returns a new panel of the class that should be used to display this
	  * card container type (e.g. BoxDisplayPanel or StreamDisplayPanel.)
	 */
	public CardContainerDisplayPanel getNewDisplay();

	/** Returns a new card browsing table panel of the appropriate class for
	  * this card container, using the given data to fill it in first.
	 */
	public CommonCardTablePanel getNewTablePanel( 
			Vector clist, CardSet cs, AnkiStreamFrame asf );

}