package ankistream;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Vector;
import java.util.TreeSet;
import java.util.Collections;
import java.util.Random;
import java.util.NoSuchElementException;
import java.util.zip.DataFormatException;
import java.io.StringReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;
import javax.swing.JOptionPane;
import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;

/** The master collection of "metacards".  Individual flash cards (Card objects)
  * are instantiated from metacards, which contain the actual display
  * information shown to the user when the cards are reviewed.
  * <br> 
  * <br> Part of the Anki Stream project, a tool for long-term flash card
  * review.  Anki Stream is distributed under the MIT/X11 free software 
  * license, a copy of which should have come with this distribution.
  * @author Bishop Wilkins
  * @version 0.1
*/ 
public class CardSet {
	
	private IXMLElement xml;	// The XML tree with the card set data.	
	
	/** Creates a new CardSet based on the specified XML tag.
	  * @param newXML "cardset" tag containing the metacard data for this card
	  *  set.
	 */
	public CardSet( IXMLElement newXML ) {
		xml = newXML;
	}
	
	/** Creates a new default CardSet.
	 */
	public CardSet( ) {
		xml = new XMLElement( "cardset" );
		xml.setAttribute( "name", "New Card Set" );
	}
	
	/** Gets the name of this CardSet.
	 */
	public String getName() {
		return xml.getAttribute( "name", "New Card Set" );
	}
	
	/** Gets the root of an XML tree containing the data for this card set.
	  * @return the root of an XML tree containing the data for this card set.
	 */
	public IXMLElement toXML( ) { return xml; }
	
	/** Sets the name of this CardSet.
	 */
	public void setName( String newval ) {
		if ( newval == null || newval.length( ) == 0 ) return;
		xml.setAttribute( "name", newval );
	}
	
	/** Gets the number of metacards in this card set.
	 * @return the number of metacards in this card set.
	 */
	public int getCardCount( ) {
		return ((Vector) xml.getChildrenNamed( "metacard" ) ).size( );
	}
	
	/** Gets the MetaCard with the given ID from this card set.
	 * @param metaCardID the ID of the MetaCard to retrieve.
	 * @return the MetaCard with the given ID, or null if no metacard could be
	 *  created.
	 */
	public MetaCard getMetaCard( int metaCardID ) {
		MetaCard output = null;
		IXMLElement mcTag = getMetaCardTag( metaCardID );
		if ( mcTag == null ) { return null; }
		String cardType = mcTag.getAttribute( "type", "unknown" );
		String className = CardTypeIndex.getClassName( cardType );
		Class c = null;
		if ( className == null ) return null;
		try {
			c = getClass( ).getClassLoader( ).loadClass( className );
			output = (MetaCard) c.newInstance( );
			output.setXML( mcTag );
		} catch ( ClassNotFoundException cnfe ) {
			return null;
		} catch ( InstantiationException ie ) {
			return null;
		} catch ( IllegalAccessException iae ) {
			return null;
		} catch ( Exception e ) {
			return null;
		}
		return output;
	}

	/** Returns the tag corresponding to the first meta card with the given ID.
	  * Or null if none could be found.
	 */
	private IXMLElement getMetaCardTag( int metaCardID ) {
		Iterator i = ((Vector) xml.getChildrenNamed( "metacard" ) ).iterator( );
		IXMLElement output = null;
		while ( i.hasNext() ) {
			output = (IXMLElement) i.next( );
			if ( metaCardID == output.getAttribute("id", -1) ) return output;
		}
		return null;
	}		
	
	/** Removes the metacard with the specified ID from the card set.
	  * Returns the IXMLTag representing the metacard removed.
	 */
	public IXMLElement removeMetaCard( int metaCardID ) {
		Iterator i = ((Vector) xml.getChildrenNamed( "metacard" ) ).iterator( );
		IXMLElement output = null;
		while ( i.hasNext() ) {
			output = (IXMLElement) i.next( );
			if ( metaCardID == output.getAttribute("id", -1) ) {
				xml.removeChild( output );
				return output;
			} 
		}
		return null;
	}

	/** Gets the next available metacard ID for this card set.
	 */
	public int getNextAvailableID( ) {
		// TODO: This could be made to return the lowest unused integer instead
		// of the next higher integer than the highest used id.
		Iterator i = ((Vector) xml.getChildrenNamed( "metacard" ) ).iterator( );
		int highestid = -1;
		IXMLElement curmc;
		int curmcid;
		while ( i.hasNext() ) {
			curmc = (IXMLElement) i.next( );
			curmcid = curmc.getAttribute("id", -1);
			if ( curmcid > highestid ) {
				highestid = curmcid;
			} 
		}
		return (highestid + 1);
	}
	
	/** Adds the given metacard to the card set.
	  * Returns the IXMLTag representing the metacard added.
	 */
	public IXMLElement addMetaCard( MetaCard mc ) {
		xml.addChild( mc.getXML() );
		return getMetaCardTag( mc.getID() );
	}
	
	/** Replaces the first metacard in the card set with the same ID as mc
	  * with mc.
	  * @return true if mc was found and replaced, false otherwise.
	 */
	public boolean replaceMetaCard( MetaCard mc ) {
		// Unfortunately this requires reassembling the entire xml tree.
		String name = getName();
		Vector v = xml.getChildren();
		xml = new XMLElement( "cardset" );
		xml.setAttribute( "name", name );

		Iterator i = v.iterator();
		boolean replaced = false;
		while (i.hasNext()) {
			IXMLElement e = (IXMLElement)i.next();
			if ( (e.getAttribute("id", -1) == mc.getID()) && !replaced ) {
				// Replace this card.
				xml.addChild(mc.getXML());
				replaced = true;
			} else {
				xml.addChild(e);
			}
		}
		
		return replaced;
	}
	
	/** Inserts mc into the card set after the card with ID afterCardID.
	  * @return true if afterCardID was found and mc added, false otherwise.
	 */
	public boolean insertMetaCardAfter( MetaCard mc, int afterCardID ) {
		// Unfortunately this requires reassembling the entire xml tree.
		String name = getName();
		Vector v = xml.getChildren();
		xml = new XMLElement( "cardset" );
		xml.setAttribute( "name", name );

		Iterator i = v.iterator();
		boolean added = false;
		while (i.hasNext()) {
			IXMLElement e = (IXMLElement)i.next();
			xml.addChild(e);
			if ( (e.getAttribute("id", -1) == afterCardID) && !added ) {
				// Insert mc here.
				xml.addChild(mc.getXML());
				added = true;
			}
		}
		
		return added;
	}
		
	/** Instantiates a new card based on the given meta card ID.
	  * @param metaCardID meta card ID to use to instantiate the card with.
	  * @return the new card.
	 */
	public StreamCard CreateCard( int metaCardID ) {
		StreamCard output = new StreamCard( );
		output.setID( metaCardID );
		return output;
	}

} // end class CardSet
