package ankistream;

import net.n3.nanoxml.IXMLElement;
import java.awt.Component;

/** <code>MetaCard</code> is the base class for all types of card.  MetaCards 
  * exist for each flash card in a CardSet, and their IDs correspond to Card IDs
  * in Boxes.  Subclasses of MetaCard will mostly be concerned with providing 
  * methods for loading data from metacards and giving display data required by 
  * their CardPanels based on their specific type of card.
  * <br>Invariant note: the xml element should always kept up to date whenever
  * your MetaCard subclass's methods update the MetaCard data in any way. 
  * The anki box framework can save the xml element to file at any time 
  * and MetaCard changes will not be saved unless their xml element is up to 
  * date when this save occurs.
  * <br> 
  * <br> Part of the Anki Stream project, a tool for long-term flash card
  * review.  Anki Stream is distributed under the MIT/X11 free software 
  * license, a copy of which should have come with this distribution.
  * @author Bishop Wilkins
  * @version 0.1
**/

public abstract class MetaCard {
	
	protected IXMLElement xml;
	
	/** Default empty constructor. */
	public MetaCard() { }
	
	/** Creates a new metacard using the given xml tag. */
	public MetaCard( IXMLElement newxml ) {
		setXML( newxml );
	}

	/** Gets the ID of this card. */
	public int getID() { return xml.getAttribute( "id", -1 ); }

	/** Gets the type of this card. */
	public String getType() {return xml.getAttribute( "type", "undefined" );}	
	
	/** Gets a short string representing this card's "question" (front side)
	  * for use in card browsing tables and such.
	 */
	public abstract String getQuestion();

	/** Gets a short string representing this card's "answer" (back side)
	  * for use in card browsing tables and such.
	 */
	public abstract String getAnswer();
	
	/** Sets the XML element of this card. */
	public void setXML( IXMLElement nXML ) {
		xml = nXML;
		initializeFromCData( );
	}
	
	/** Gets this card as an XML tag tree. */
	public IXMLElement getXML( ) { return xml; }
	
	/** Sets the "type" attribute in the xml element to the appropriate value:
	  * a string representing this type of card.
	  * @param newval new type of the card.
	 */
	public void setType( String newval ) {
		xml.setAttribute( "type", newval );
	}
	
	/** Sets the ID of this card. */
	public void setID( int newID ) { 
		xml.setAttribute( "id", String.valueOf( newID ) );
	}
	
	/** Sets the contents of this card. */
	public void setContents( String newcontents ) {
		xml.setContent( newcontents );
	}
	
	/** Initializes the fields of this MetaCard based on the contents of the
	  * tag in the xml file.  These may be, for example, a series of 
	  * tab-delimited fields, or whatever else you want to use to retrieve and
	  * store your data.
	 */
	public abstract void initializeFromCData( );
	
	/** Returns a display panel for this metacard. */
	public abstract MetaCardDisplayPanel getDisplayPanel( );

	/** Returns an editing panel for editing/adding this type of metacard. */
	public abstract MetaCardDisplayPanel getEditPanel( );

	/** Returns an editing panel for editing/adding this type of metacard. 
	  * With the specified component to be highlighted after editing. */
	public abstract MetaCardDisplayPanel getEditPanel( Component c );
	
	// TODO: temp workaround for getting font sizes to edit panel without
	// FontController.
	public abstract MetaCardDisplayPanel getEditPanel( Component c, Stream s);

} // end abstract class MetaCard
