package ankistream;

import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;

/** A wrapper for the "preferences" tag of a anki stream data file, containing
  * user preferences for the anki stream application.
  * <br> 
  * <br> Part of the Anki Stream project, a tool for long-term flash card
  * review.  Anki Stream is distributed under the MIT/X11 free software 
  * license, a copy of which should have come with this distribution.
  * @author Bishop Wilkins
  * @version 0.1
*/ 
public class Preferences {
	
	IXMLElement xml;
	
	/** Creates a new preferences object with default values.
	 */
	public Preferences( ) {
		xml = new XMLElement( "preferences" );
		setValue( "firststartup", "no" );
		setValue( "defaultstreamfile", "AnkiStream.xml" );
		setValue( "opendefaultstreamfileonstartup", "yes" );
		setValue( "language", "0" );
	}
	
	/** Creates a "deep" copy of the specified Preferences object.
	 */	
	public Preferences( Preferences sourceprefs ) {
		xml = new XMLElement( "preferences" );
		setValue(	"defaultstreamfile", 
					sourceprefs.getValue( "defaultstreamfile" ) );
		setValue( 	"opendefaultstreamfileonstartup", 
					sourceprefs.getValue( "opendefaultstreamfileonstartup" ) );
		setValue( "language", sourceprefs.getValue("language") );
	}
	
	/** Creates a new preferences object with the values specified in the given
	  * XML tag.
	  * @param iXML a "preferences" tag with values to use for this preferences
	  *  object.
	 */
	public Preferences( IXMLElement iXML ) {
		xml = iXML;		
	}
	
	public String getDefaultStreamFile( ) {
		IXMLElement tag = xml.getFirstChildNamed( "defaultstreamfile" );
		if ( tag == null ) {
			setValue( "defaultstreamfile", "AnkiStream.xml" );
			return "AnkiStream.xml";
		}
		return tag.getAttribute( "value", "AnkiStream.xml" );		
	}
	
	public IXMLElement toXML( ) {
		return xml;
	}
	
	// Sets the "value" attribute of the given tag to the given value.
	// Creates the tag and/or attribute if they do not exist.
	private void setValue( String tagname, String value ) {
		IXMLElement tag = xml.getFirstChildNamed( tagname );
		if ( tag == null ) {
			tag = new XMLElement( tagname );
			xml.addChild( tag );
		}
		tag.setAttribute( "value", value );		
	}
	
	// Gets the literal "value" attribute of the specified tag.
	private String getValue( String tagname ) {
		IXMLElement tag = xml.getFirstChildNamed( tagname );
		return tag.getAttribute( "value", "" );
	}
	
	public boolean getOpenDefaultStreamFileOnStartup( ) {
		IXMLElement tag = xml.getFirstChildNamed("opendefaultstreamfileonstartup");
		if ( tag == null ) {
			setValue( "opendefaultstreamfileonstartup", "yes" );
			return true;
		}
		if ( tag.getAttribute( "value", "yes" ).compareTo( "yes" ) == 0 ) {
			return true;
		} else {
			return false;
		}
	}

	public boolean getFirstStartup( ) {
		IXMLElement tag = xml.getFirstChildNamed("firststartup");
		if ( tag == null ) {
			setValue( "firststartup", "no" );
			return true;
		}
		if ( tag.getAttribute( "value", "yes" ).compareTo( "yes" ) == 0 ) {
			return true;
		} else {
			return false;
		}
	}

	public int getLanguage( ) {
		IXMLElement tag = xml.getFirstChildNamed("language");
		if ( tag == null ) {
			setValue( "language", "0" );
			return 0;
		}

		int rval = 0;
		try {
			rval = Integer.parseInt(tag.getAttribute( "value", "0" ));
		} catch (Exception e) {
			rval = 0;
		}

		return rval;
	}

	public void setLanguage( int newval ) {
		setValue( "language", String.valueOf(newval) );
	}
	
	public void setFirstStartup( boolean newval ) {
		String setval = "yes";
		if ( !newval ) { setval = "no"; }
		setValue( "firststartup", setval );
	}

	public void setDefaultStreamFile( String newval ) {
		setValue( "defaultstreamfile", newval );
	}

	public void setOpenDefaultStreamFileOnStartup( boolean newval ) {
		String setval = "yes";
		if ( !newval ) { setval = "no"; }
		setValue( "opendefaultstreamfileonstartup", setval );
	}
	
} // end class Preferences
