package ankistream;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Vector;
import java.util.TreeSet;
import java.util.Collections;
import java.util.Random;
import java.util.Date;
import java.util.Properties;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.zip.DataFormatException;
import java.text.ParseException;
import java.io.StringReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.FileNotFoundException;
import java.io.UnsupportedEncodingException;
import javax.swing.JOptionPane;
import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;

/** A stream of flash cards arranged in order from the soonest to be in need of
  * review in front to the most well-known in back.  The stream can return the
  * best card to review next, place it back into the stream upstream when it
  * is declared wrong or right, and provides some other functions for searching
  * through the cards and rearranging them.
  * <br>
  * <br> Internally, the stream is an ordered vector of (date, xml tag pointer)
  * pairs for quicker reordering of the cards according to nextreviewtime.  The
  * vector is always kept ordered, but the underlying stream tag is only
  * reassembled in order when toXML() is called.
  * <br>
  * <br> Invariant: the internal vector is kept ordered between public function
  * calls.
  * <br>
  * <br> Part of the Anki Stream project, a tool for long-term flash card
  * review.  Anki Stream is distributed under the MIT/X11 free software 
  * license, a copy of which should have come with this distribution.
  * @author Bishop Wilkins
  * @version 0.1
*/ 
public class Stream implements CardContainer {
	
	private class Entry implements Comparable {
		public java.util.Date date;
		public IXMLElement xml;
		public int compareTo(Object o) {
			if (!(o instanceof Entry)) {
				throw new ClassCastException("Cant compare non-streamcards.");
			}
			Entry e = (Entry)o;
			return date.compareTo(e.date);
		}
		
		public boolean equals(Object o) {
			if ( !(o instanceof Entry) ) {
				throw new ClassCastException("Cant compare non-streamcards.");
			}
			Entry e = (Entry)o;
			return date.equals(e.date);
		}
	}
	
	private Vector undoStack;
	
	private Vector entries;
	
	private IXMLElement xml;	// The most recently assembled XML tree with 
								// the stream data.	
	private boolean xmlOutOfDate; // does xml need to be reassembled (ordered)?
	private Random random;		// For shuffling cards.
	
	// Millisecond time tick to review until, -1 if no review is underway.
	private long reviewUntil;

	// Tracking stats for the last reviewed card:
	private StreamCard.StreamCardStats lastCardStats;
	
	private long defaultReviewLength = 1800000; // 1/2 hour in millis
	private int right;			// Current review stats.
	private int wrong;
	
	private int lifetimerights;
	private int lifetimereviews;
	
	private long firstcarddate;
	
	/** Creates a new empty stream.
	 */
	public Stream( ) {
		undoStack = new Vector();
		lastCardStats = null;
		lifetimerights = 0;
		lifetimereviews = 0;
		firstcarddate = Long.MAX_VALUE;
		xml = new XMLElement("stream");
		createEntriesFromXML(xml);
		xmlOutOfDate = true;
		random = new Random( System.currentTimeMillis( ) );
		reviewUntil = -1;
		right = 0;
		wrong = 0;		
	}
		
	/** Creates a new Stream using the data in the given XML tag tree.
	  * @param newXML the root ("stream" tag) of an XML tree containing the
	  *  data for this Stream. 
	 */
	public Stream( IXMLElement newXML ) {
		
		undoStack = new Vector();
		lastCardStats = null;

		lifetimerights = 0;
		lifetimereviews = 0;
		firstcarddate = Long.MAX_VALUE;
		
		xml = newXML;
		createEntriesFromXML(xml);
		xmlOutOfDate = true;

		random = new Random( System.currentTimeMillis( ) );

		reviewUntil = -1;
		right = 0;
		wrong = 0;
	}
	
	/** Creates a new Stream with default attributes.
	 */
	public Stream( String cardSetFile ) {
		
		undoStack = new Vector();
		lastCardStats = null;

		lifetimerights = 0;
		lifetimereviews = 0;
		firstcarddate = Long.MAX_VALUE;
		
		xml = new XMLElement( "stream" );
		xml.setAttribute( "name", "New Stream" );
		xml.setAttribute( "cardsetfile", AnkiStreamFrame.defaultDataFile );
		xml.setAttribute(	"secondspercard", 
							String.valueOf(DEFAULT_SECONDS_PER_CARD) );
		createEntriesFromXML(xml);
		xmlOutOfDate = false;
		
		random = new Random( System.currentTimeMillis( ) );

		reviewUntil = -1;
		right = 0;
		wrong = 0;
	}

	private void createEntriesFromXML( IXMLElement newXML ) {
		entries = new Vector(); // Clear entries and rebuild with add() calls.
		Vector tags = xml.getChildrenNamed( "card" );
		Iterator i = tags.iterator();
		while (i.hasNext()) {
			IXMLElement x = (IXMLElement)i.next();
			StreamCard sc = new StreamCard( x );
			lifetimerights += sc.getLifetimeRights();
			lifetimereviews += sc.getLifetimeReviews();
			long cdate = StreamCard.parseTime(sc.getCreationDate());
			if (cdate < firstcarddate) { firstcarddate = cdate; }
			add( sc, false );
		}
		sortAndSpreadDuplicateTimes();
	}
	
	/** Gets the root of an XML tree containing the data for this stream.
	  * @return the root of an XML tree containing the data for this stream.
	 */
	public IXMLElement toXML( ) { 
		rebuildXML();
		return xml; 
	}
	
	/** Gets the name of this stream.
	  * @return the name of this stream
	 */
	public String getName( ) {
		return xml.getAttribute( "name", "Unnamed Stream" );
	}
	
	/** Sets the name of this stream.
	 */
	public void setName( String newname ) {
		if ( newname == null || newname.length( ) == 0 ) return;
		xml.setAttribute( "name", newname );
	}
	
	/** Gets the seconds per card that this card container uses for reviews.
	 */
	public int getSecondsPerCard( ) {
		return xml.getAttribute("secondspercard",DEFAULT_SECONDS_PER_CARD);
	}

	/** Sets the seconds per card that this card container uses for reviews.
	 */
	public void setSecondsPerCard( int newval ) {
		if (newval < 0) {newval = 0;}
		xml.setAttribute("secondspercard", String.valueOf(newval));
	}

	/** Gets the font size that this card container uses for review questions.
	 */
	public float getQuestionFontSize( ) {
//		return Float.parseFloat( xml.getAttribute( "questionfontsize",
//				String.valueOf(FontController.getUnicodeFont().getSize()) ) );		
		return Float.parseFloat(xml.getAttribute("questionfontsize", "14.0"));		
	}

	/** Sets the font size that this card container uses for review questions.
	 */
	public void setQuestionFontSize( int newval ) {
		xml.setAttribute( "questionfontsize", String.valueOf( newval ) );	
	}

	/** Gets the font size that this card container uses for review answers.
	 */
	public float getAnswerFontSize( ) {
//		return Float.parseFloat( xml.getAttribute( "answerfontsize",
//				String.valueOf(FontController.getUnicodeFont().getSize()) ) );
		return Float.parseFloat(xml.getAttribute("answerfontsize", "14.0"));		
	}

	/** Sets the font size that this card container uses for review answers.
	 */
	public void setAnswerFontSize( int newval ) {
		xml.setAttribute( "answerfontsize", String.valueOf( newval ));		
	}

	/** Gets whether a card's review time should be randomized somewhat after
	  * its recommended review time goes above 1 week.
	 */
	public boolean getRandomizeReviewTimes( ) {
		return Boolean.parseBoolean(
				xml.getAttribute("randomizereviewtimes", "false"));
	}

	/** Sets whether a card's review time should be randomized somewhat after
	  * its recommended review time goes above 1 week.
	 */
	public void setRandomizeReviewTimes( boolean newval ) {
		xml.setAttribute( "randomizereviewtimes", String.valueOf( newval ));		
	}

	/** Gets whether card review time spans are capped at the max review time.
	 */
	public boolean getUseMaxReviewTime( ) {
		return Boolean.parseBoolean(
				xml.getAttribute("usemaxreviewtime", "true"));
	}

	/** Sets whether card review time spans are capped at the max review time.
	 */
	public void setUseMaxReviewTime( boolean newval ) {
		xml.setAttribute( "usemaxreviewtime", String.valueOf( newval ));		
	}

	/** Gets the max review span per card in days.
	 */
	public int getMaxReviewTime( ) {
		return Integer.parseInt( xml.getAttribute( "maxreviewtime", "60" ) );
	}

	/** Gets the max review span per card in millis.
	 */
	public long getMaxReviewTimeMillis( ) {
		int i = Integer.parseInt( xml.getAttribute( "maxreviewtime", "60" ) ); 
		return (i*(long)86400000);
	}

	/** Sets the max review span per card in days.
	 */
	public void setMaxReviewTime( int newval ) {
		xml.setAttribute( "maxreviewtime", String.valueOf(newval) );		
	}
	
	/** Gets whether the stream is autosaved according to the autosave setting.
	 */
	public boolean getUseAutoSaveEvery( ) {
		return Boolean.parseBoolean(
				xml.getAttribute("useautosaveevery", "true"));
	}

	/** Sets whether the stream is autosaved according to the autosave setting.
	 */
	public void setUseAutoSaveEvery( boolean newval ) {
		xml.setAttribute( "useautosaveevery", String.valueOf( newval ));		
	}

	/** Gets the number of cards to autosave after.
	 */
	public int getAutoSaveEvery( ) {
		return Integer.parseInt(xml.getAttribute( "autosaveevery", "100" ));
	}

	/** Sets the number of cards to autosave after.
	 */
	public void setAutoSaveEvery( int newval ) {
		xml.setAttribute( "autosaveevery", String.valueOf(newval) );		
	}

	/** Gets whether stream should be autosaved after each review.
	 */
	public boolean getAutoSaveAfterReview( ) {
		return Boolean.parseBoolean(
				xml.getAttribute("autosaveafterreview", "true"));
	}

	/** Sets whether stream should be autosaved after each review.
	 */
	public void setAutoSaveAfterReview( boolean newval ) {
		xml.setAttribute( "autosaveafterreview", String.valueOf( newval ));		
	}
	
	/** Returns the end time of the current review in milliseconds, or the end
	  * time of the review were one to start right now.
	 */
	public long getReviewEndTime() {
		long rval = reviewUntil;
		if (rval == -1) {
			rval = new Date().getTime();
			rval += getDefaultReviewLength();
		}
		return rval;
	}
	
	/**  Gets the default review length in milliseconds.
	 */
	public long getDefaultReviewLength() {
		return defaultReviewLength;
	}

	/**  Sets the default review length in milliseconds.
	 */
	public void setDefaultReviewLength( long newlength ) {
		defaultReviewLength = newlength;
	}
	
	/**  Gets the total number of lifetime correct answers. */
	public long getLifetimeRights() { return lifetimerights; }

	/**  Gets the total number of lifetime reviews. */
	public long getLifetimeReviews() { return lifetimereviews; }
	
	/** Gets the millisecond date of this stream's earliest card's creation,
	 *  or -1 if there are no cards in the stream. */
	public long getFirstCardDate() { 
		if (firstcarddate == Long.MAX_VALUE) { return -1; }
		return firstcarddate; 
	}

	/** Gets the name of the data file containing the card set for this stream.
	 */
	public String getCardSetFile( ) {
		return xml.getAttribute( "cardsetfile", 
			AnkiStreamFrame.defaultDataFile );
	}

	/** Sets the name of the data file containing the card set for this stream.
	 */
	public void setCardSetFile( String newfile ) {
		if ( newfile == null || newfile.length( ) == 0 ) return;
		xml.setAttribute( "cardsetfile", newfile );
	}
	
	/** Gets the zoom level this stream was last saved at.
	 */
	public String getInitialZoom( ) {
		return xml.getAttribute( "initialzoom", "auto" );
	}
	
	/** Sets the zoom level this stream was last saved at.
	 */
	public void setInitialZoom( String newval ) {
		xml.setAttribute( "initialzoom", newval );
	}

	/** Gets the min review span per card in minutes.
	 */
	public int getMinReviewTime( ) {
		return Integer.parseInt( xml.getAttribute( "minreviewtime", "5" ) );
	}

	/** Gets the min review span per card in millis.
	 */
	public long getMinReviewTimeMillis( ) {
		int i = Integer.parseInt( xml.getAttribute( "minreviewtime", "5" ) ); 
		return (i*(long)60000);
	}
	
	/** Sets the min review span per card in minutes.
	 */
	public void setMinReviewTime( int newval ) {
		xml.setAttribute( "minreviewtime", String.valueOf(newval) );		
	}

	/** Gets the correct card review timesspan multiplier.
	 */
	public float getIncreaseSpanBy( ) {
		return Float.parseFloat( xml.getAttribute( "increasespanby", "6.0" ) );
	}

	/** Sets the correct card review timesspan multiplier.
	 */
	public void setIncreaseSpanBy( float newval ) {
		xml.setAttribute( "increasespanby", String.valueOf(newval) );		
	}

	/** Gets the correct card review timesspan multiplier for penalized cards.
	 */
	public float getWorstPenaltyMultiplier( ) {
		return Float.parseFloat( 
				xml.getAttribute( "worstpenaltymultiplier", "2.0" ) );
	}

	/** Sets the correct card review timesspan multiplier for penalized cards.
	 */
	public void setWorstPenaltyMultiplier( float newval ) {
		xml.setAttribute( "worstpenaltymultiplier", String.valueOf(newval) );		
	}

	/** Gets whether often wrong cards are penalized with the above multiplier.
	 */
	public boolean getPenalizeHardCards( ) {
		return Boolean.parseBoolean( 
				xml.getAttribute("penalizehardcards", "true"));
	}

	/** Sets whether often wrong cards are penalized with the above multiplier.
	 */
	public void setPenalizeHardCards( boolean newval ) {
		xml.setAttribute( "penalizehardcards", String.valueOf( newval ));		
	}
	
	/** Rebuilds the underlying xml tree according to the order of the cards in
	  * the entries vector, if necessary.
	 */
	private void rebuildXML() {
		
		if (!xmlOutOfDate) { return; }
		
		IXMLElement oldXML = xml;
		xml = new XMLElement( "stream" );

		// Copy attributes.
		Properties oldAttributes = oldXML.getAttributes();
		Enumeration en = oldAttributes.propertyNames();
		while (en.hasMoreElements()) {
			String name = (String)en.nextElement();
			xml.setAttribute(name, oldAttributes.getProperty(name));
		}
		
		// Copy elements.
		Iterator i = entries.iterator();
		while ( i.hasNext() ) {
			Entry e = (Entry)i.next();
			xml.addChild(e.xml);
		}

		xmlOutOfDate = false;
	}
	
	/** Returns the number of instances in this stream of the card with the
	  * given id.
	 */
	public int instanceCount( int cardid ) {
		int c = 0;
		Iterator i = entries.iterator();
		StreamCard curcard;
		while (i.hasNext()) {
			curcard = new StreamCard( ((Entry)i.next()).xml );
			if ( curcard.getID() == cardid ) c++;
		}
		return c;
	}
	
	public StreamCard getFirstCardWithID( int id ) {
		Entry e = getEntryByCardID( id );
		return new StreamCard( e.xml );
	}
		
	/** Gets ids of all cards containing the given text string in the question 
	  * or answer.
	  * @param searchstring the String to search for in the cards.
	  * @param cset card set whose metacards' text we should search through.
	  * @return A vector of ints representing the IDs of the cards containing
	  * the queriedstring.
	*/
	public Vector findCardsContainingString(	String searchstring,
												CardSet cset ) { 
		//TODO:
		// Search each card for the specified target string.
		Vector rval = new Vector();
		
		Iterator i = iterator();
		while ( i.hasNext() ) {
			StreamCard sc = (StreamCard)i.next();
			MetaCard mc = cset.getMetaCard(sc.getID());
			String text = mc.getQuestion()+" "+mc.getAnswer();
			if (text.indexOf(searchstring) != -1) {
				rval.add(new Integer(sc.getID()));
			}
		}

		return rval;
	}
			
	/** Gets the number of cards in the stream.
	 */
	public int getCardCount( ) {		
		return entries.size( );
	}
	
	/** Gets the index of the first card in the stream after the given time.
	  * Or -1 if no such card exists. */
	public int getFirstCardIndexAfterTime( long target ) {
		if ( entries.size() == 0 ) { return -1; }
		return getCardCountBefore(target);
	}

	/** Gets the index of the first card in the stream before the given time.
	  * Or -1 if no such card exists. */
	public int getFirstCardIndexBeforeTime( long target ) {
		return getCardCountBefore(target) - 1;
	}

	/** Gets the number of cards in the stream before the specified time,
	  * using a binary search for the target time.
	 */
	public int getCardCountBefore( long target ) {
		
		// special case: no cards
		if (entries.size() == 0) { return 0; }
		
		// special case: target time < first card time (no cards before this)
		if (((Entry)entries.firstElement()).date.getTime() > target) {
			return 0;
		}
		
		// special case: target time > last card time
		if ((((Entry)entries.lastElement()).date.getTime()) < target) {
			return getCardCount();
		}
		
		int low = 0;
		int high = entries.size() - 1;
		int mid = high / 2;
		long curtime = ((Entry)entries.get(mid)).date.getTime();
		while ( (high-low) > 1 ) {
			curtime = ((Entry)entries.get(mid)).date.getTime();
			if (curtime == target) { break; }
			if (curtime < target) {
				// Search upstream.
				low = mid;
				mid += ((high-low)/2);
			} else {
				// Search downstream.				
				high = mid;
				mid -= ((high-low)/2);
			}
		}
				
		if ((curtime < target)) { mid++; }
	
		return mid;
	}

	/** Adds the specified card to the stream.
	  * @param newCard the card to be added
	 */
	public void add( CommonCard newCard ) {
		if ( newCard instanceof Card ) {
			newCard = ((Card)newCard).toStreamCard();
		}
		add((StreamCard)newCard, true);
	}

	/** Adds the specified stream card to the stream.
	  * @param newCard the card to be added
	 */
	public void add( StreamCard newCard ) {
		add((StreamCard)newCard, true);
	}

	/** Adds the specified card to the stream.
	  * @param newCard the card to be added
	  * @param keepInvariant set false to not do re-sorting and extra checking
	  * to maintain the invariant (used internally for batch adds, not 
	  * available to external callers)
	 */
	private void add( StreamCard newCard, boolean keepInvariant ) {

		Entry e = new Entry();
		e.xml = newCard.toXML( );
		
		// Assign date if this is a new card.
		if (newCard.getCreationDate().equalsIgnoreCase("now") ) {
			long birthDate = new Date().getTime();
			newCard.setCreationDate(StreamCard.formatTime(birthDate));
			newCard.setBestNextReviewDate(StreamCard.formatTime(birthDate));
			e.xml = newCard.toXML();
		}

		try {
			e.date = 
				StreamCard.dateform.parse(newCard.getBestNextReviewDate());
		} catch ( ParseException pe ) {
			e.date = new Date();
		}
		
		lifetimerights += newCard.getLifetimeRights();
		lifetimereviews += newCard.getLifetimeReviews();
		long cdate = StreamCard.parseTime(newCard.getCreationDate());
		if (cdate < firstcarddate) { firstcarddate = cdate; }
		
		entries.add(e);
		if (keepInvariant) {sortAndSpreadDuplicateTimes();}
		xmlOutOfDate = true;
	}

	// Sorts the cards inside the stream by best next review time.  If any
	// have the exact same review time (to the second) they are spread apart
	// by one second.
	private void sortAndSpreadDuplicateTimes() {
		Collections.sort(entries);
		
		Iterator i = entries.iterator();
		if ( !i.hasNext() ) {return;}

		// Don't allow two cards with the exact same time (space close times
		// by 1000 millis.
		Date lastdate = ((Entry)i.next()).date;
		while (i.hasNext()) {
			Entry e = (Entry)i.next();
			if (e.date.compareTo(lastdate) <= 0) {
				StreamCard sc = new StreamCard( e.xml );
				long t = lastdate.getTime() + 1000;
				sc.setBestNextReviewDate(StreamCard.formatTime(t));
				e.date = new Date(t);
				e.xml = sc.toXML();
			}
			lastdate = e.date;
		}
	}

	/** Removes the specified card from the stream.
	  * @param cardID ID of card to remove.
	  * @return true if the card was found and removed, false otherwise.
	 */
	public boolean remove( int cardID ) {
		Entry e = getEntryByCardID( cardID );
		if ( e == null ) return false;

		StreamCard sc = new StreamCard(e.xml);
		lifetimerights -= sc.getLifetimeRights();
		lifetimereviews -= sc.getLifetimeReviews();

		entries.remove(e);
		Collections.sort(entries);
		xmlOutOfDate = true;
		
		long cdate = StreamCard.parseTime(sc.getCreationDate());
		if (cdate==firstcarddate) { calculateNewFirstCard(); }

		return true;
	}
	
	private void calculateNewFirstCard() {
		firstcarddate = Long.MAX_VALUE;
		Vector tags = xml.getChildrenNamed( "card" );
		Iterator i = tags.iterator();
		while (i.hasNext()) {
			IXMLElement x = (IXMLElement)i.next();
			StreamCard sc = new StreamCard( x );
			long cdate = StreamCard.parseTime(sc.getCreationDate());
			if (cdate < firstcarddate) { firstcarddate = cdate; }
		}		
	}
	
	private class StreamIterator implements Iterator {
		
		private Iterator i;
		
		public StreamIterator() {
			i = entries.iterator();
		}
		
		public boolean hasNext() {
			return i.hasNext();
		}
		
		public Object next() {
			StreamCard sc = new StreamCard( ((Entry)i.next()).xml );
			return sc;
		}
		
		public void remove() {
			// (not implemented)
			return;
		}
		
	}

	private class ReverseStreamIterator implements Iterator {
		
		private Entry e;
		private int curindex;
		
		public ReverseStreamIterator() {
			curindex = entries.size()-1;
			if (curindex >= 0 ) { e = (Entry)entries.get(curindex); }
			else {e = null;}
		}
		
		public boolean hasNext() {
			return (e != null);
		}
		
		public Object next() {
			StreamCard sc = new StreamCard( e.xml );
			curindex--;
			if (curindex >= 0 ) { e = (Entry)entries.get(curindex); }
			else {e = null;}
			return sc;
		}
		
		public void remove() {
			// (not implemented)
			return;
		}
		
	}

	/** Gets an iterator that iterates through each of the cards in this
	  * stream in order.  If the stream is mutated, the returned iterator's
	  * behavior becomes unpredictable.
	 */
	public Iterator iterator() {
		return new StreamIterator();
	}

	/** Gets a reverse iterator that iterates through each of the cards in this
	  * stream in reverse order, starting from the end.  If the stream is 
	  * mutated, the returned iterator's behavior becomes unpredictable.
	 */
	public Iterator reverseIterator() {
		return new ReverseStreamIterator();
	}

	private Entry getEntryByCardID( int cardID ) {
		Iterator i = entries.iterator();
		while ( i.hasNext() ) {
			Entry e = (Entry)i.next();
			IXMLElement x = e.xml;
			if ( x.getAttribute("id", "-1").compareToIgnoreCase( 
					String.valueOf(cardID) ) == 0 ) return e;
		}
		return null;
	}
		
	/** Begins reviewing every card in the stream whose "bestreviewtime" is
	  * earlier than the default review end time.
	 */
	public void beginReview( ) {
		beginReview( true, defaultReviewLength );
	}

	/** Begins reviewing every card in the stream whose "bestreviewtime" is
	  * earlier than the specified review end time (specified as number of
	  * milliseconds after the current time).
	 */
	public void beginReview( long timespan ) {
		beginReview( true, timespan );
	}

	/** Begins reviewing every card in the stream whose "bestreviewtime" is
	  * earlier than the specified review end time (specified as number of
	  * milliseconds after the current time).
	  * @param endreviewfirst if true, end the current review if one is 
	  * underway and start a new one; if false, just continue if a review is
	  * underway.
	 */
	public void beginReview( boolean endreviewfirst, long timespan ) {
		
		if ( !endreviewfirst && reviewUntil != -1 ) {
			// Requested no end review and we are in a review.  Just continue.
			return;
		}

		if ( endreviewfirst ) {

			endReview( );
			
		}

		undoStack.clear();

		// reset score
		setRight( 0 );
		setWrong( 0 );
		reviewUntil = new Date().getTime() + timespan;
	}
	
	/** Ends any review currently in session.
	 */
	public void endReview( ) {
		reviewUntil = -1;
	}
	
	/** Gets the number of cards right so far this review session.
	 * @return the number of cards right so far this review session.
	 */
	public int getRight() {
		return right;
	}

	/** Sets the number of cards right so far this review session.
	  * @param newval the new number of cards right.
	 */
	public void setRight( int newval ) {
		right = newval;
	}
	
	/** Gets the number of cards wrong so far this review session.
	 * @return the number of cards wrong so far this review session.
	 */
	public int getWrong() {
		return wrong;
	}

	/** Sets the number of cards right so far this review session.
	  * @param newval the new number of cards right.
	 */
	public void setWrong( int newval ) {
		wrong = newval;
	}

	/** Gets the number of cards reviewed so far this review session.
	 * @return the number of cards reviewed so far this review session.
	 */
	public int getReviewed() {
		return (getRight( ) + getWrong( ));
	}

	/** Gets a string of the form "[right] / [total] ([percent right] %)" 
	  * describing the status of the current review session.
 	  * @return the status string.
	 */
	public String getReviewStatusString() {
		Float fval1 = Float.valueOf(getRight());
		Float fval2 = Float.valueOf(getReviewed());
		int percentCorrect = new Float 
			((fval1.floatValue() / fval2.floatValue())*100.0).intValue();
		String s =
			getRight() + " / " + getReviewed() + " (" +  percentCorrect + " %)";
		return s;
	}

	/** Gets the ID of the card under review (the front card in the stream).
	  * @return the ID of the card under review, or -1 if there is no review
	  *  in session.
	 */
	public int getCurrentCardID( ) {
		if (reviewUntil == -1 ) {return -1;}
		IXMLElement c = getCurrentCardXML( );
		if ( c == null ) return -1; 
		return c.getAttribute( "id", -1 );
	}
		
	/** Returns a copy of the card under review, or null if there is not a 
	  * review session underway.
	 */
	public StreamCard getCurrentCard( ) {
		if (reviewUntil == -1 ) {return null;}
		IXMLElement c = getCurrentCardXML( );
		if ( c == null ) return null; 
		return new StreamCard( c );
	}

	private IXMLElement getCurrentCardXML( ) {
		if ( entries.size() == 0 ) return null;
		return ((Entry)entries.get(0)).xml;
	}

	/** Returns the best review date for the card under review in millis. 
	  * Or -1 if there is no review underway. */
	public long getCurrentCardBestReviewDate() {
		StreamCard sc = getCurrentCard();
		if ( sc == null ) { return -1; }
		else {
			return StreamCard.parseTime(sc.getBestNextReviewDate());
		}
	}

	/** Returns a copy of the frontmost card in the stream, whether or not
	  * a review is under way.  Or returns null if the stream is empty.
	 */
	private StreamCard getFrontCard( ) {
		if ( entries.size() == 0 ) return null;
		return new StreamCard(((Entry)entries.get(0)).xml);
	}

	/** Returns the best review date for frontmost card in the stream in 
	  * millis, whether or not a review is underway.  Returns current time if
	  * the stream is empty. */
	public long getFrontCardBestReviewDate() {
		StreamCard sc = getFrontCard();
		if ( sc == null ) { return new Date().getTime(); }
		else {
			return StreamCard.parseTime(sc.getBestNextReviewDate());
		}
	}
	
	/** Returns a copy of the rearmost card in the stream, whether or not
	  * a review is under way.  Or returns null if the stream is empty.
	 */
	private StreamCard getBackCard( ) {
		if ( entries.size() == 0 ) return null;
		return new StreamCard(((Entry)entries.lastElement()).xml);
	}
	
	/** Returns the best review date of the rearmost card in the stream in
	  * millis, whether or not a review is underway.  Returns current time if
	  * the stream is empty.
	 */
	public long getBackCardBestReviewDate() {
		StreamCard sc = getBackCard();
		if ( sc == null ) { return new Date().getTime(); }
		else {
			return StreamCard.parseTime(sc.getBestNextReviewDate());
		}
	}
	
	/** Returns the stats for the last card reviewed.
	  * or null if no card has been reviewed yet this session.
	 */
	public StreamCard.StreamCardStats getLastCardStats() {
		return lastCardStats;
	}

	/** Called when the current card under review is declared right.
	 */
	public void right( ) {
		
		UndoData ud = new UndoData();
		ud.right = true;
		ud.oldStats = lastCardStats;

		// Temporarily remove card from the stream and update card stats.
		StreamCard sc = getFrontCard();
		remove(sc.getID());
		ud.oldCard = new StreamCard( sc.toXML() );
		lastCardStats = sc.right(this);

		// Put the card back into the stream at a new place.
		add(sc);
		xmlOutOfDate = true;

		// Increment the score
		setRight( getRight( ) + 1 );

		// Update undo data.
		undoStack.insertElementAt(ud, 0);
		
		// End the review if it's time.
		endReviewIfDone();
	}
		
	/** Called when the current card under review is declared wrong.
	 */
	public void wrong( ) {

		UndoData ud = new UndoData();
		ud.right = false;
		ud.oldStats = lastCardStats;

		// Temporarily remove card from the stream and update card stats.
		StreamCard sc = getFrontCard();
		remove(sc.getID());
		ud.oldCard = new StreamCard( sc.toXML() );
		lastCardStats = sc.wrong(this);

		// Put the card back into the stream at a new place.
		add(sc);
		xmlOutOfDate = true;

		// Increment the score
		setWrong( getWrong( ) + 1 );

		// Update undo data.
		undoStack.insertElementAt(ud, 0);

		// End the review if it's time.
		endReviewIfDone();
	}

	/** Ends the review if the next card's bestreviewtime is later than the
	  * review end time.
	 */
	private void endReviewIfDone() {
		long nextCardTime = reviewUntil+1;
		try {
			nextCardTime =
				StreamCard.dateform.parse(
						getCurrentCard().getBestNextReviewDate()).getTime();
		} catch (ParseException pe) {
			
		}
		if ( nextCardTime > reviewUntil ) {
			endReview();
		}
	}
	
	/** Returns true if the last action on this card container can be undone.
	 */
	public boolean canUndo() {
		return (undoStack.size() > 0);
	}
	
	/** Undoes the effects of the last Right( ) or Wrong( ) call.
	  * Returns true if the undo was successful.
	 */
	public boolean Undo( ) {
		
		if (undoStack.size() < 1) { return false; }

		UndoData ud = (UndoData)undoStack.remove(0);
		if ( ud.right ) { right--; }
		else { wrong--; }
		lastCardStats = ud.oldStats;
		reviewUntil = ud.oldReviewUntil;
		remove(ud.oldCard.getID());
		add(ud.oldCard);
		return true;
	}

	/** Holds data about what tasks one "undo" action should perform.
	 */
	private class UndoData {
		public UndoData() {
			oldReviewUntil = reviewUntil;
			oldStats = null;
		}
		public StreamCard oldCard;		// Which card needs to be undone.
		public boolean right;			// Was it right or wrong?
		public long oldReviewUntil;		// Old reviewUntil value.
		public StreamCard.StreamCardStats oldStats;
	}

	/** Returns a new panel of the class that should be used to display this
	  * card container type (e.g. BoxDisplayPanel or StreamDisplayPanel.)
	 */
	public CardContainerDisplayPanel getNewDisplay() {
		return new StreamDisplayPanel();
	}

	/** Returns a new card browsing table panel of the appropriate class for
	  * this card container, using the given data to fill it in first.
	 */
	public CommonCardTablePanel getNewTablePanel( 
			Vector clist, CardSet cs, AnkiStreamFrame asf ) {
		return new StreamCardTablePanel( clist, cs, asf );
	}
	
} // end class Stream
