package ankistream;

import java.util.Date;
import java.util.Random;
import java.util.Calendar;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.text.DecimalFormat;

import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;

/** <code>Card</code> represents one flash card in a stream.  Cards have IDs
  * that correspond to the MetaCards which actually contain their display
  * information, and a set of statistics about review history.
  * <br> They are compared for ordering by their bestnextreviewdates.
  * <br> 
  * <br> Part of the Anki Stream project, a tool for long-term flash card
  * review.  Anki Stream is distributed under the MIT/X11 free software 
  * license, a copy of which should have come with this distribution.
  * @author Bishop Wilkins
  * @version 0.1
**/

// TODO: refactor to Card after old Card class can be removed
public class StreamCard implements CommonCard, Comparable {
	
	IXMLElement xml;
	
	public static final Random random = new Random(new Date().getTime());
	
	public static final SimpleDateFormat dateform = 
		new SimpleDateFormat( "yyyy.MM.dd hh:mm:ss aa" );
	
	public static final long MAX_REVIEW_LENGTH_MILLIS = 
		Long.parseLong("5184000000"); // 60 days

	public static final long REVIEW_LENGTH_AFTER_MISS = 
		Long.parseLong("300000"); // 5 min
	
	public static final long RANDOMIZE_AFTER = 604800000; // 1 week in millis
	public static final double RANDOMIZE_BY = .1; // by 1/10th the time span
	
	/** Compares cards based on their best next review time.
	  * @return a negative integer, zero, or a positive integer as this object
	  * is less than, equal to, or greater than the specified object.
	 */
	public int compareTo( Object o ) {
		if (!(o instanceof StreamCard)) {
			throw new ClassCastException("Can't compare StreamCard to " 
					+ o.getClass().getName() );
		}
		StreamCard sc = (StreamCard)o;
		return dateform.format(getBestNextReviewDate()).compareTo(
					dateform.format(sc.getBestNextReviewDate()));
	}
	
	/** Creates a new Card based on the specified XML "card" tag.
	  * @param iXML an xml "card" tag containing data about a card.
	 */
	public StreamCard( IXMLElement iXML ) {
		xml = new XMLElement( "card" );
		setID( iXML.getAttribute("id", -1 ) );
		setCreationDate( iXML.getAttribute("creationdate", "now") );
		setLastReviewDate(iXML.getAttribute("lastreviewdate", "never reviewed"));
		setLastReviewSpan(iXML.getAttribute("lastreviewspan","none"));
		setBestNextReviewDate( iXML.getAttribute("bestnextreviewdate", "now") );
		setLastHundredHistory( iXML.getAttribute("lasthundredhistory", "") );
		setLastWrongDate( iXML.getAttribute("lastwrongdate", "never missed") );
		setLifetimeRights( iXML.getAttribute("lifetimerights", 0) );
		setLifetimeWrongs( iXML.getAttribute("lifetimewrongs", 0) );
		setAverageForgetPeriod(
				iXML.getAttribute("averageforgetperiod","never missed"));
	}
	
	/** Creates a new uninitialized card. */
	public StreamCard( ) {
		String birthtime = getTimeString( );
		xml = new XMLElement( "card" );
		setCreationDate( birthtime );
		setLastReviewDate( "never reviewed" );
		setLastReviewSpan( "none" );
		setBestNextReviewDate( birthtime );
		setLastHundredHistory( "" );
		setLastWrongDate( "never missed" );
		setLifetimeRights( 0 );
		setLifetimeWrongs( 0 );
		setAverageForgetPeriod( "never missed" );
	}
	
	/** Gets the xml element of this card.  Should only be called by the
	  * Anki Stream framework when a new card is being created. */
	public IXMLElement toXML( ) { return xml; }
	
	/** Gets the ID of this card. */
	public int getID() { return xml.getAttribute("id", -1); }

	/** Gets the creation time of this card. */
	public String getCreationDate() { 
		return xml.getAttribute("creationdate", "now"); 
	}

	/** Gets the last review time of this card. */
	public String getLastReviewDate( ) { 
		return xml.getAttribute("lastreviewdate", "never reviewed"); 
	}
	
	/** Gets the last review time span of this card. */
	public String getLastReviewSpan( ) {
		return xml.getAttribute("lastreviewspan", "never reviewed");
	}
	
	/** Gets the best next review time of this card. */
	public String getBestNextReviewDate( ) { 
		return xml.getAttribute("bestnextreviewdate", "now"); 
	}
	
	/** Gets a string of the form xx/yy (zz%) [  streak: +/- aa] 
	  * describing the history of this card.
	**/
	public String getHistoryString( ) {
		// TODO: refactor this block into a utility function?  It's copy-pasted
		// in a few places...
		String rval = getLifetimeRights() + " / " + getLifetimeReviews();
		Float fval1 = Float.valueOf(getLifetimeRights());
		Float fval2 = Float.valueOf(getLifetimeReviews());
		int percentCorrect = new Float 
			((fval1.floatValue() / fval2.floatValue())*100.0).intValue();
		rval += " (" + percentCorrect + " %)";
		int streak = getStreak();
		if ( streak < -1 || streak > 1 ) {
			if ( streak < 0 ) {
				rval += Strings.get(115) + streak;
			} else {
				rval += Strings.get(116) + streak;
			}
		}
		return rval;
	}

	/** Gets the history string of right/wrong bits for last 100 reviews. */
	public String getLastHundredHistory( ) {
		return xml.getAttribute("lasthundredhistory", ""); 
	}
	
	/** Gets the number of rights in the last 100 reviews. */
	public int getRightsInLastHundred( ) {
		String s = getLastHundredHistory();
		int rights = 0;
		for ( int i = 0; i < s.length(); i++ ) {
			if ( s.charAt(i) == '1' ) { rights++; }
		}
		return rights;
	}

	/** Gets the number of wrongs in the last 100 reviews. */
	public int getWrongsInLastHundred( ) {
		String s = getLastHundredHistory();
		int wrongs = 0;
		for ( int i = 0; i < s.length(); i++ ) {
			if ( s.charAt(i) == '0' ) { wrongs++; }
		}
		return wrongs;
	}

	/** Gets the last time this card was missed in review. */
	public String getLastWrongDate( ) { 
		return xml.getAttribute("lastwrongdate", "now"); 
	}

	/** Returns a positive number representing the number of times this card
	  * has been correctly identified in review, or a negative number
	  * representing the number of times it has been missed. */
	public int getStreak( ) {
		int index = getLastHundredHistory().length();
		if (index == 0) return 0; // Never reviewed.
		int streak = 1;
		index--;
		char streakchar = getLastHundredHistory().charAt(index);
		while (index > 0 ) {
			index--;
			if ( getLastHundredHistory().charAt(index) == streakchar ) streak++;
			else break;
		}
		if (streakchar == '0') streak *= -1; // It was a losing streak.
		return streak; 
	}

	/** Gets the number of times this card has been correctly identified. */
	public int getLifetimeRights( ) { 
		return xml.getAttribute("lifetimerights", 0); 
	}

	/** Gets the number of times this card has been missed. */
	public int getLifetimeWrongs( ) {
		return xml.getAttribute("lifetimewrongs", 0); 
	}	
	
	/** Gets the total number of times this card has been reviewed. */
	public int getLifetimeReviews( ) { 
		return getLifetimeRights() + getLifetimeWrongs();
	}

	/** Gets the average timespan this card has been forgotten in. */
	public String getAverageForgetPeriod( ) { 
		return xml.getAttribute("averageforgetperiod", "never missed"); 
	}	
	
	
	/** Sets the ID of this card. */
	public void setID( int newID ) { 
		xml.setAttribute( "id", String.valueOf( newID ) );
	}

	/** Sets the creation time of this card. */
	public void setCreationDate( String val ) { 
		xml.setAttribute( "creationdate", val );
	}

	/** Sets the last review time of this card. */
	public void setLastReviewDate( String val ) { 
		xml.setAttribute( "lastreviewdate", val );
	}

	/** Sets the last review time span of this card. */
	public void setLastReviewSpan( String val ) {
		xml.setAttribute( "lastreviewspan", val );
	}

	/** Sets the best next review time of this card. */
	public void setBestNextReviewDate( String val ) { 
		xml.setAttribute( "bestnextreviewdate", val );
	}

	/** Sets the history string of right/wrong bits for last 100 reviews. */
	public void setLastHundredHistory( String val ) { 
		xml.setAttribute( "lasthundredhistory", val );
	}

	/** Sets the last time this card was missed in review. */
	public void setLastWrongDate( String val ) { 
		xml.setAttribute( "lastwrongdate", val );
	}

	/** Sets the number of times this card has been correctly identified. */
	public void setLifetimeRights( int val ) { 
		xml.setAttribute( "lifetimerights", String.valueOf( val ) );
	}

	/** Sets the number of times this card has been missed. */
	public void setLifetimeWrongs( int val ) { 
		xml.setAttribute( "lifetimewrongs", String.valueOf( val ) );
	}

	/** Sets the average time this card is forgotten in. */
	public void setAverageForgetPeriod( String val ) { 
		xml.setAttribute( "averageforgetperiod", val );
	}
	
	/** Does all increments and internal bookkeeping necessary for when this
	  * card is answered right during a review session.
	 */
	public StreamCardStats right(Stream s) {
		
		long maxReviewTime=s.getMaxReviewTime()*(long)86400000; //Days->Millis
		long minReviewTime = s.getMinReviewTime()*(long)60000; // Min->millis.
		
		Date curDate, prevLastReviewDate, prevBestReviewDate;
		curDate = new Date();
		try {prevLastReviewDate = dateform.parse(getLastReviewDate());}
		catch (ParseException pe) {prevLastReviewDate = new Date();}
		try {prevBestReviewDate = dateform.parse(getBestNextReviewDate());}
		catch (ParseException pe) {prevBestReviewDate = new Date();}

		setLastReviewDate( dateform.format( curDate ) );
		setLifetimeRights(getLifetimeRights()+1);
		addCharToLastHundredHistory( '1' );
		
		// Calculate best next review date.
		long lastTimeSpan = 
			prevBestReviewDate.getTime() - prevLastReviewDate.getTime();
		lastTimeSpan = Math.max(lastTimeSpan, minReviewTime);
		// Actually use the lastreviewspan attribute if available.
		if (formatTimeSpan(getLastReviewSpan()) != 0) {
			lastTimeSpan = formatTimeSpan(getLastReviewSpan());
		}
		StreamCardStats rval = new StreamCardStats();
		rval.lastReviewTime = prevLastReviewDate.getTime();
		rval.lastReviewSpan = StreamCard.formatTimeSpanShortForm(lastTimeSpan);
		rval.lastReviewSpanLong = lastTimeSpan;
		long bestNextTimeSpanEstimate = 
			(long)((float)lastTimeSpan*s.getIncreaseSpanBy());
		// If the option is checked,
		// add penalty ("failure drag") for cards that are missed a lot.  Cards 
		// missed 1/2 the time or more should have the maximum penalty.
		if (s.getPenalizeHardCards()) {
			long worstNextTimeSpanEstimate = 
				(long)((float)lastTimeSpan*s.getWorstPenaltyMultiplier());
			long worstBestDifference = 
				bestNextTimeSpanEstimate - worstNextTimeSpanEstimate;
			long doubleWrongs =
				Math.min((getWrongsInLastHundred()*2),
						(getLastHundredHistory().length()));
			float failureDragRatio = 
				(float)doubleWrongs/(float)(getLastHundredHistory().length());
			long failureDragSpan = 
				(long)((float)worstBestDifference*failureDragRatio);
			bestNextTimeSpanEstimate -= failureDragSpan;
		}
		long nextTimeSpan =
			Math.min(bestNextTimeSpanEstimate, maxReviewTime);
		rval.nextReviewSpan = StreamCard.formatTimeSpanShortForm(nextTimeSpan);
		rval.nextReviewSpanLong = nextTimeSpan;
		rval.record = getHistoryString();
		Date bestNextReviewDate = new Date(curDate.getTime()+nextTimeSpan);
		// Also randomize the next review date if requested.
		if (s.getRandomizeReviewTimes() && (nextTimeSpan > RANDOMIZE_AFTER)) {
			int randomRange = (int)((double)nextTimeSpan * RANDOMIZE_BY);
			int randomizeBy = random.nextInt( randomRange );
			randomizeBy -= ( randomRange / 2 ); // Allow negative numbers.
			bestNextReviewDate = 
				new Date(bestNextReviewDate.getTime()+randomizeBy);
			rval.randomOffset = randomizeBy;
		}
		setBestNextReviewDate(dateform.format(bestNextReviewDate));
		setLastReviewSpan( formatTimeSpan(nextTimeSpan) );
		rval.newReviewTime = bestNextReviewDate.getTime();
		return rval;
	}
	
	/** Does all increments and internal bookkeeping necessary for when this
	  * card is answered wrong during a review session.
	 */
	public StreamCardStats wrong( Stream s ) {
		
		long minReviewTime=s.getMinReviewTime()*(long)60000; //Minutes->millis
		
		Date curDate,prevLastReviewDate,prevLastWrongDate,prevBestReviewDate;
		
		curDate = new Date();
		try {prevLastReviewDate = dateform.parse(getLastReviewDate());}
		catch (ParseException pe) {prevLastReviewDate = new Date();}
		try {prevLastWrongDate = dateform.parse(getLastWrongDate());}
		catch (ParseException pe) {prevLastWrongDate = new Date();}
		try {prevBestReviewDate = dateform.parse(getBestNextReviewDate());}
		catch (ParseException pe) {prevBestReviewDate = new Date();}
		
		setLastReviewDate( dateform.format( curDate ) );
		setLastWrongDate( dateform.format( curDate ) );
		setLifetimeWrongs(getLifetimeWrongs()+1);
		addCharToLastHundredHistory( '0' );
		
		// Calculate best next review date.
		long lastTimeSpan = 
			prevBestReviewDate.getTime() - prevLastReviewDate.getTime();
		lastTimeSpan = Math.max(lastTimeSpan, minReviewTime);
		// Actually use the lastreviewspan attribute if available.
		if (formatTimeSpan(getLastReviewSpan()) != 0) {
			lastTimeSpan = formatTimeSpan(getLastReviewSpan());
		}
		StreamCardStats rval = new StreamCardStats();
		rval.lastReviewTime = prevLastReviewDate.getTime();
		rval.lastReviewSpan = StreamCard.formatTimeSpanShortForm(lastTimeSpan);
		rval.lastReviewSpanLong = lastTimeSpan;
		long nextTimeSpan = minReviewTime;
		setLastReviewSpan( formatTimeSpan(nextTimeSpan) );
		Date bestNextReviewDate = new Date(curDate.getTime()+nextTimeSpan);
		setBestNextReviewDate(dateform.format(bestNextReviewDate));
		rval.newReviewTime = bestNextReviewDate.getTime();
		rval.nextReviewSpan = StreamCard.formatTimeSpanShortForm(nextTimeSpan);
		rval.nextReviewSpanLong = nextTimeSpan;
		rval.newReviewTime = bestNextReviewDate.getTime();
		rval.record = getHistoryString();
		
		// Calculate new averageforgetperiod.
		long curForgetPeriod = curDate.getTime()-prevLastReviewDate.getTime();
		long avgForgetPeriod = formatTimeSpan(getAverageForgetPeriod());
		avgForgetPeriod *= getLifetimeReviews() - 1;
		avgForgetPeriod += curForgetPeriod;
		avgForgetPeriod /= getLifetimeReviews();
		setAverageForgetPeriod(formatTimeSpan(avgForgetPeriod));		
		
		return rval;
	}
	
	/** Adds the given character (should be 1 or 0) to the history as the most
	  *  recent review result. */
	private void addCharToLastHundredHistory( char c ) {
		String newval = getLastHundredHistory();
		newval += c;
		if (newval.length() > 100) {
			newval = newval.substring(newval.length()-100, newval.length());
		}
		this.setLastHundredHistory(newval);
	}
	
	/** Converts to a Card. */
	public Card toCard() {
				
		Card c = new Card( );
		c.setID( getID() );
		c.setCreationDate(convertDate(getCreationDate()));
		c.setLastReviewDate(convertDate(getLastReviewDate()));
		c.setLastWrongDate(convertDate(getLastWrongDate()));
		c.setLifetimeRights(getLifetimeRights());
		c.setLifetimeWrongs(getLifetimeWrongs());
		
		return c;
	}

	private String convertDate(String cardformat) {
		try {
			return Card.dateform.format(StreamCard.dateform.parse(cardformat));
		} catch ( ParseException pe ) {
			return cardformat;
		}
	}
	
	/** Returns the value in milliseconds of the given Stream Card style
	  * formated yyyy.mm.dd.ss date string, or 0 if string could not be parsed.
	 */
	public static long parseTime( String s ) {
		try {
			return dateform.parse(s).getTime();
		} catch (ParseException pe) {
			return 0;
		}
	}
	
	/** Returns a Stream Card style yyyy.mm.dd.ss date string corresponding to
	  * the given time in milliseconds.
	 */
	public static String formatTime( long t ) {
		return dateform.format( new Date(t) );
	}
	
	/** Gets the time in yyyy.mm.dd.ss format. */
	public static String getTimeString() {
		return dateform.format( new Date( ) );
	}
	
	/** Converts from a millisecond time span to a string formatted time span.*/
	public static String formatTimeSpan( long span ) {
		// Format is "[Y years, ][X days, ][Z hours, ][A minutes, ]B seconds".
		String rval = "";
		long years = 0, days = 0, hours = 0, minutes = 0, seconds = span / 1000;
		
		// Calculate the units.
		if ( seconds > 60 ) {
			minutes = (seconds / 60);
			seconds -= (minutes*60);
		}
		if ( minutes > 60 ) {
			hours = (minutes / 60);
			minutes -= (hours*60);
		}
		if ( hours > 24 ) {
			days = (hours / 24);
			hours -= (days*24);
		}
		if ( days > 365 ) {
			years = (days / 365);
			days -= (days * 365);
		}
		
		// Assemble the string.
		if ( years > 0 ) {
			rval += years + " years, ";
		}
		if ( days > 0 ) {
			rval += days + " days, ";
		}
		if ( hours > 0 ) {
			rval += hours + " hours, ";
		}
		if ( minutes > 0 ) {
			rval += minutes + " minutes, ";
		}
		rval += seconds + " seconds";
		
		return rval;
	}

	/** Converts from a millisecond time span to a string formatted time span
	  * with just the highest unit shown.  */
	public static String formatTimeSpanShortForm( long span ) {
		span /= 1000;
		double dspan = (double)span;
		//TODO: merge with duplicate code in streamdisplaypanel?
		String rval;
		DecimalFormat df = new DecimalFormat("#0.0");
		df.setMaximumFractionDigits(1);
		if ( span < 2*60 ) {// < 2 minutes
			df = new DecimalFormat("#0");
			rval = df.format(dspan) + Strings.get(117);
		} else if ( span < (2*3600) ) { // < 2 hours
			rval = df.format(dspan / 60.0) + Strings.get(118);
		} else if ( span < (2*86400) ) { // < 48 hours
			rval = df.format(dspan / 3600.0) + Strings.get(119);
		} else if ( span < (2*2592000) ) { // < 2 months
			rval = df.format(dspan / 86400.0) + Strings.get(120);
		} else if ( span < (4*2592000) ) { // < 4 months
			rval = df.format(dspan / 604800.0) + Strings.get(121);
		} else if ( span < (4*77760000) ) { // < 2 years
			rval = df.format(dspan / 2419200.0) + Strings.get(122);
		} else {
			rval = df.format(dspan / 31536000.0) + Strings.get(123);
		}
		return rval;
	}

	/** Converts from a string formatted time span to a millisecond time span.
	  * Returns 0 if there was a problem during the conversion. */
	public static long formatTimeSpan( String span ) {
		long rval = 0;
		String[] tokens = span.split(" ");
		int imax = tokens.length;
		int i = imax-1;
		// Parse from "[Y years, ][X days, ][Z hours, ][A minutes, ]B seconds".
		try {
			while ( i > 0 ) {
				if (tokens[i].compareToIgnoreCase("seconds") == 0) {
					rval += (Long.parseLong(tokens[i-1])) * 1000;

				} else if (tokens[i].compareToIgnoreCase("minutes,") == 0) {
					rval += (Long.parseLong(tokens[i-1])) * 60000;

				} else if (tokens[i].compareToIgnoreCase("hours,") == 0) {
					rval += (Long.parseLong(tokens[i-1])) * 3600000;

				} else if (tokens[i].compareToIgnoreCase("days,") == 0) {
					rval += (Long.parseLong(tokens[i-1])) * 86400000;

				} else if (tokens[i].compareToIgnoreCase("years,") == 0) {
					rval += (Long.parseLong(tokens[i-1])) * 
								Long.parseLong("31536000000");
				}
				i -= 2;
			}			
		} catch (NumberFormatException nfe) {
			return 0;
		}
		return rval;
	}

	/** Gets the date in yyyy.mm.dd format. */
	public String getDateString() {
		return dateform.format( new Date( ) );
	}
	
	/** Returns a string suitable for "Last seen: XXX" notes.  Examples:
	  * Yesterday 12:00 PM; Today 3:00 AM; Friday 8:30 PM; Dec 12; Aug 5 2009
	 */
	public static String formatRelativeShortTime( long t ) {

		String rval = "unknown";

		Date now = new Date();
		Date target = new Date(t);
		SimpleDateFormat sdf = new SimpleDateFormat( "yyyy.MM.dd" );
		String nowstr = sdf.format(now);
		String tarstr = sdf.format(target);
		int nowint, tarint;
				
		if (nowstr.equalsIgnoreCase(tarstr)) {
			// Today:
			sdf = new SimpleDateFormat("hh:mm aa");
			rval = Strings.get(124) + sdf.format(target);
		} else {
			sdf = new SimpleDateFormat("yyyy");
			nowstr = sdf.format(now);
			tarstr = sdf.format(target);
			if ( nowstr.equalsIgnoreCase(tarstr)) {
				// Same year:
				sdf = new SimpleDateFormat("DDD"); // Day in year
				nowstr = sdf.format(now);
				tarstr = sdf.format(target);
				nowint = Integer.parseInt(nowstr);
				tarint = Integer.parseInt(tarstr);
				if ((nowint-tarint) == 1) {
					// Yesterday
					sdf = new SimpleDateFormat("hh:mm aa");
					rval = Strings.get(125) + sdf.format(target);
				} else if ((nowint-tarint) == -1) {
					// Tomorrow
					sdf = new SimpleDateFormat("hh:mm aa");
					rval = Strings.get(126) + sdf.format(target);
				} else {
					long timediff = Math.abs(now.getTime()-target.getTime());
					if (timediff>518400000) {
						// Different week.
						sdf = new SimpleDateFormat("MMM dd");
						rval = sdf.format(target);
					} else {
						// Same week.
						sdf = new SimpleDateFormat("EEE hh:mm aa");
						rval = sdf.format(target);
					}
				}
			} else {
				// Different year.
				long timediff = Math.abs(now.getTime()-target.getTime());
				if (timediff>518400000) {
					// Different week.
					sdf = new SimpleDateFormat("MMM dd yyyy");
					rval = sdf.format(target);
				} else {
					// Same week.
					sdf = new SimpleDateFormat("EEE hh:mm aa");
					rval = sdf.format(target);
				}
			}
		}
		
		return rval;
	}
	
	/** A class that holds statistics about how often a stream card has been
	  * reviewed and its last and next estimated best review times.  Returned
	  * by right() and wrong().
	 */
	public class StreamCardStats {
		public long lastReviewSpanLong;
		public long nextReviewSpanLong;
		public String lastReviewSpan;
		public long lastReviewTime;
		public long newReviewTime;
		public String nextReviewSpan;
		public String record;
		public int randomOffset;
		
		public StreamCardStats() {
			lastReviewSpan = "";
			lastReviewTime = 0;
			newReviewTime = 0;
			nextReviewSpan = "";
			record = "";
			randomOffset = 0;
		}
	}
	
} // end class StreamCard
