package ankistream;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.Color;
import java.awt.Point;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GridLayout;
import java.awt.image.BufferedImage;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent; 
import java.util.Date;
import java.util.Vector;
import java.util.Iterator;
import java.util.Random;
import java.text.DecimalFormat;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JCheckBox;
import javax.swing.BorderFactory;
import javax.swing.JOptionPane;
import javax.swing.SwingConstants;
import javax.swing.Timer;
import javax.swing.event.ChangeListener; 
import javax.swing.event.ChangeEvent;
import javax.swing.event.MouseInputAdapter;

import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;

/** A panel displaying a anki stream and its cards visually.  Shown on the
  * right hand side of the main frame of the Anki Stream application.
  * <br> 
  * <br> Part of the Anki Stream project, a tool for long-term flash card
  * review.  Anki Stream is distributed under the MIT/X11 free software 
  * license, a copy of which should have come with this distribution.
  * @author Bishop Wilkins
  * @version 0.1
*/ 
public class StreamDisplayPanel 
	extends JPanel implements CardContainerDisplayPanel {
	
	private StreamRenderPanel srp;
	private LastCardPanel lcp;
	private boolean lcpVisible;
	
	public StreamDisplayPanel() {
		setLayout( new BorderLayout( ) );
		lcpVisible = false;
		lcp = new LastCardPanel();
		srp = new StreamRenderPanel();
		add(srp, BorderLayout.CENTER);
		setPreferredSize( new Dimension( 360, 480 ) );
	}
	
	// Shows the last card panel if there is a last card reviewed to show,
	// or hides it if not.
	private void toggleLastCardPanel() {
		
		StreamCard.StreamCardStats scs = srp.stream.getLastCardStats();

		// Relabel the last card reviewed stats panel.
		if (scs != null) {
			lcp.redraw(scs, srp.stream);
		}

		// Toggle the panel layout if necessary.
		boolean lcpVisibleNow = (scs != null);
		if (lcpVisible == lcpVisibleNow) {
			return; // Not necessary.
		} else {
			lcpVisible = lcpVisibleNow;
		}
		
		if (lcpVisible) {
			remove(srp);
			add(lcp, BorderLayout.SOUTH);
			add(srp, BorderLayout.CENTER);
		} else {
			remove(srp);
			remove(lcp);
			add(srp, BorderLayout.CENTER);
		}
		
		validate();
		showCurrentAndLastReviewed();
	}

	/** Begins invoking specified AnkiStreamFrame's onSelectionChanged()
	  * method whenever a new compartment is selected/deselected. 
	 */
	public void notifyMeOnSelection( AnkiStreamFrame newowner ) {
		srp.notifyMeOnSelection(newowner);
	}

	/** Auto-adjusts the slider to show both the next card for review
	  * and the last reviewed card.  Or shows the front of the stream if there
	  * is no last reviewed card.
	 */
	public void showCurrentAndLastReviewed() {
		toggleLastCardPanel();
		srp.showCurrentAndLastReviewed();
	}

	/** Redraws the graphical representation of the box.
	 */
	public void redraw( ) {
		toggleLastCardPanel();
		srp.redraw();
	} // end redraw
	
	/** Called whenever the user clicks the mouse on the display panel.
	  * Tests whether the user clicked on a card and if so selects that
	  * card area unless a review is underway.
	  * @param x x position the mouse was clicked at. Relative to this panel.
	  * @param y y position the mouse was clicked at. Relative to this panel.
	 */
	public void onClick( int x, int y ) {
		srp.onClick(x, y);
	}

	// TODO: some kind of get and set selected card range mechanismm
	/** Returns the id of the selected card, or -1 if no card is selected.
	 */
	public int getSelectedCardID() {
		return srp.getSelectedCardID();
	}
	
	/** Gets a set of cards around and including the selected card to browse.
	 */
	public Vector getCardRange() {
		return srp.getCardRange();
	}
	
	public void setCardContainer( CardContainer cc ) {
		srp.setCardContainer(cc);
	}

	/** Gives this renderer an interface to access its card container.
	 */	
	public void setStream( Stream newstream ) {		
		srp.setStream(newstream);
	} // end setStream

	public boolean getAutoZoom( ) {
		return srp.getAutoZoom();
	}

	/** Sets the Zoom slider without redrawing the panel. */
	public void setZoomSliderNoRedraw(int newval) {
		srp.setZoomSliderNoRedraw(newval);
	}

	public void setZoomSlider(int newval) {
		srp.setZoomSlider(newval);
	}
	
	public void setAutoZoom( boolean newval ) {
		srp.setAutoZoom( newval );
	}	
	
	public void adjustRangeToZoomSlider() {
		srp.adjustRangeToZoomSlider();
	}

	public void adjustZoomSliderToRange() {
		srp.adjustZoomSliderToRange();
	}
	
	public void initiateMove( ) {
		srp.initiateMove();
	}
	
	public void selectCardWithID( int id ) {
		srp.selectCardWithID(id);
	}
	
	private class StreamRenderPanel 
		extends JPanel implements CardContainerDisplayPanel {
		
		boolean redrawOnSliderChange;
		
		private Stream stream;
		private BufferedImage streamimage; // Double buffer for drawing the stream.
		
		private JSlider zoomSlider;
		private JLabel zoomLabel;
		private JCheckBox autoZoom;
	
		private AnkiStreamFrame ownerframe;
		
		private Timer refreshtimer;
		
		private long minDate;
		private long maxDate;
		
		private boolean isUserDragSelecting;
		private boolean isUserMovingCards;
		private Point selectionStart;
		private Point selectionEnd;
		private Vector selectedCards;
		
		private long earliestdtime = 0;
		private long latestdtime = 0;
		private double secondheight = 0;		
		
		public StreamRenderPanel( ) {

			redrawOnSliderChange = true;

			minDate = getCurrentTime() - 7200000;
			maxDate = getCurrentTime() + 7200000;
			
			Color bgcolor = new Color( 196, 196, 255 );
			ownerframe = null;
			setBackground( bgcolor );
			setBorder( BorderFactory.createTitledBorder( "" ) );

			isUserDragSelecting = false;
			isUserMovingCards = false;
			selectionStart = null;
			selectionEnd = null;
			selectedCards = null;

			// Lay out the panel.
			setLayout( new BorderLayout( ) );
			// Lay out the west panel.
			JPanel westpanel = new JPanel(new BorderLayout());
			westpanel.setBackground(bgcolor);
			zoomLabel = new JLabel(Strings.get(141));
			zoomLabel.setBackground( bgcolor );
			westpanel.add(zoomLabel, BorderLayout.NORTH);
			zoomSlider = new JSlider(JSlider.VERTICAL, 1, 100, 90);
			zoomSlider.setBackground( bgcolor );
			zoomSlider.setLabelTable(zoomSlider.createStandardLabels(10, 10));
			zoomSlider.setPaintTicks(true);
			zoomSlider.setPaintLabels(true);
			zoomSlider.addChangeListener( 
					new ChangeListener() {
						public void stateChanged( ChangeEvent ce ) { 
							onSliderChanged();
						}
					} );
			westpanel.add(zoomSlider, BorderLayout.CENTER);
			autoZoom = new JCheckBox(Strings.get(142));
			autoZoom.setBackground(bgcolor);
			autoZoom.setSelected(true);
			autoZoom.addActionListener(
					new ActionListener() {
						public void actionPerformed( ActionEvent ae ) { 
							if (autoZoom.isSelected()) {
								stream.setInitialZoom("auto");
								ownerframe.setCardContainerChangesMade(true);
								showCurrentAndLastReviewed();
								redraw();
							}
						}
					} );
			westpanel.add(autoZoom, BorderLayout.SOUTH);
			this.add(westpanel, BorderLayout.WEST);
			
			addMouseListener(
				new MouseAdapter() {
					public void mouseClicked( MouseEvent e ) {
						if (isUserMovingCards) {
							onMoveDestinationSelected( e.getX(), e.getY() );
						}
					}
					public void mousePressed( MouseEvent e ) {
						if (isUserMovingCards) {return;}
						if (stream.getCurrentCard() == null) {
							isUserDragSelecting = true;
							selectionStart = new Point(e.getX(), e.getY());
							selectionEnd = new Point(e.getX(), e.getY());
							repaint();
						}
					}
					public void mouseExited( MouseEvent e ) {
						if (stream.getCurrentCard() == null
								&& isUserDragSelecting ) {
							isUserDragSelecting = false;
							selectionStart = null;
							selectionEnd = null;
							repaint();
						}
					}
					public void mouseReleased( MouseEvent e ) {
						if (isUserMovingCards) {return;}
						if (stream.getCurrentCard() == null
								&& isUserDragSelecting ) {
							isUserDragSelecting = false;
							selectionEnd = new Point(e.getX(), e.getY());
							onClick(e.getX(),e.getY());
							repaint();
						}
					}
				}
			);

			addMouseMotionListener(
				new MouseInputAdapter() {
					public void mouseDragged( MouseEvent e ) {
						if (stream.getCurrentCard() == null
								&& isUserDragSelecting ) {
							selectionEnd = new Point(e.getX(), e.getY());
							repaint();
						}
					}
				}
			);
			
			// Refresh the display every few seconds to see the cards stream down.
			ActionListener timerListener = new ActionListener( ) {
				public void actionPerformed(ActionEvent evt) {
					redraw( );
				}
			};
			refreshtimer = new Timer( 5000, timerListener );
			refreshtimer.start();		
		}
		
		/** Begins invoking specified AnkiStreamFrame's onSelectionChanged()
		  * method whenever a new compartment is selected/deselected. 
		 */
		public void notifyMeOnSelection( AnkiStreamFrame newowner ) {
			ownerframe = newowner;
		}
		
		/** Paints this panel.
		 */
		public void paintComponent( Graphics g ) {
	
			super.paintComponent( g ); 
	
			if ( streamimage != null  ) {
				g.drawImage( streamimage, 0, 0, null );

				// Draw card selection box if user is selecting.
				if ( stream.getCurrentCard() == null && isUserDragSelecting
						&& selectionStart != null && selectionEnd != null ) {
					int x = Math.min( selectionStart.x, selectionEnd.x );
					int y = Math.min( selectionStart.y, selectionEnd.y );
					int w = Math.abs( selectionStart.x - selectionEnd.x );
					int h = Math.abs( selectionStart.y - selectionEnd.y );
					g.setColor(Color.WHITE);
					g.drawRect(x, y, w, h);
				}
			}
		}
		
		public boolean getAutoZoom( ) {
			return autoZoom.isSelected();
		}

		public void setAutoZoom( boolean newval ) {
			autoZoom.setSelected( newval );
		}

		public long getCurrentCardDate( ) {
			return stream.getCurrentCardBestReviewDate();
		}
		
		public long getCurrentTime( ) {
			return new Date().getTime();
		}
		
		public long getReviewEndTime( ) {
			return stream.getReviewEndTime();
		}
		
		public long getBackCardDate( ) {
			return stream.getBackCardBestReviewDate();
		}
		
		/** Returns either the current time or the earliest review card time,
		  * whichever is earlier.
		 */
		private long getFirstDisplayTime( ) {
			long earliesttime = stream.getFrontCardBestReviewDate();
			earliesttime = Math.min(earliesttime, getCurrentTime());
			return earliesttime;
		}
		
		/**
		 */
		private long getLastDisplayTime( ) {
			long latesttime = getBackCardDate();
			latesttime = Math.max(latesttime, getReviewEndTime());
			return latesttime;
		}
				
		/** Auto-adjusts the slider to show both the next card for review
		  * and the last reviewed card.  Or shows the front of the stream if there
		  * is no last reviewed card.
		 */
		public void showCurrentAndLastReviewed() {
			
			// Return if user does not want to auto-adjust view.
			if (!autoZoom.isSelected()) { 
				redraw();
				return; 
			}
			
			// Calculate desired time for display
			long firsttime = this.getFirstDisplayTime();
			
			StreamCard.StreamCardStats scs = stream.getLastCardStats();
			long lasttime = (scs == null) ?  
					stream.getBackCardBestReviewDate() : scs.newReviewTime;
			lasttime = Math.max(lasttime, getReviewEndTime());
			// At least four minutes' timespan:
			if ((lasttime-firsttime) < 240000) {lasttime = firsttime + 240000;}
	
			// Add padding to either end.
			minDate = firsttime;
			maxDate = lasttime;
			minDate -= getEndPadding();
			maxDate += getEndPadding();
	
			adjustZoomSliderToRange();
		}
		
		private void onSliderChanged( ) {
			if (!autoZoom.isSelected()) {
				stream.setInitialZoom(String.valueOf(zoomSlider.getValue()));
				ownerframe.setCardContainerChangesMade(true);
			}
			if (redrawOnSliderChange) {
				adjustRangeToZoomSlider();
				redraw( ); 
			}
		}
		
		// Gets the max amount of padding in milliseconds that should be used
		// displaying either the front or end of the stream, based on the 
		// current zoom level.
		private long getEndPadding( ) {
			double drange = (double)(maxDate - minDate);
			return (long)(drange*.01);
		}
		
		// Calculate the total time available for display in seconds.
		private long getTotalTimeSpan( ) {
			long earliesttime = getFirstDisplayTime();
			long latesttime = getLastDisplayTime();
			return latesttime - earliesttime;			
		}
		
		/** Sets the zoom slider to match as closely as possible the current
		  * display range size.
		 */
		public void adjustZoomSliderToRange() {
			
			double drange = (double)(maxDate - minDate);
			
			// Calculate the slider position based on the desiredrange.
			// (Reverse of calculation performed in redraw().
			// Split slider bar into 3 segments by slider range for smoother zoom:
			// 67-100 first 5% of span
			// 34-66  next 20% of span
			// 1-33   next 75% of span
			// TODO: Should instead use calculus or something to figure out
			// how to smoothly scale the zoom slider increments.
			long mintimespan = 240000; // 4 minute minimum scale
			double actualtotal = (double)(getTotalTimeSpan()-mintimespan);
			double span_1 = actualtotal * .01;
			double inc_1 = span_1 / 33.0;
			double span_2 = actualtotal * .09;
			double inc_2 = span_2 / 33.0;
			double span_3 = actualtotal * .90;
			double inc_3 = span_3 / 33.0;
			int sliderval = 100;
			while ((sliderval > 66) & (drange > 0)) {
				sliderval-= 1;
				drange -= inc_1;
			}
			while ((sliderval > 33) & (drange > 0)) {
				sliderval-= 1;
				drange -= inc_2;
			}
			while ((sliderval > 1) & (drange > 0)) {
				sliderval-= 1;
				drange -= inc_3;
			}
			
			setZoomSliderNoRedraw(sliderval);
		}
		
		/** Sets the Zoom slider without redrawing the panel. */
		public void setZoomSliderNoRedraw(int newval) {
			redrawOnSliderChange = false;
			zoomSlider.setValue(newval);
			redrawOnSliderChange = true;
		}
		
		public void setZoomSlider(int newval) {
			zoomSlider.setValue(newval);
		}
		
		/** Sets the display range to match as closely as possible the current
		  * value of the zoom slider.
		 */
		public void adjustRangeToZoomSlider() {

			// Metrics for time span calculations.
			long currenttime = getCurrentTime();
			long earliesttime = getFirstDisplayTime( );
			long latesttime = getLastDisplayTime( );
			long maxtimespan = getTotalTimeSpan( );

			// Calculate the desired display span and center.
			// Split slider bar into 3 segments by slider range for smoother 
			// zoom:
			// 67-100 first 5% of span
			// 34-66  next 20% of span
			// 1-33   next 75% of span
			// TODO: Should instead use calculus or something to figure out
			// how to smoothly scale the zoom slider increments.
			long mintimespan = 240000; // 4 minute minimum scale
			double slider = zoomSlider.getMaximum()-zoomSlider.getValue()+1;
			double actualtotal = (double)(maxtimespan-mintimespan);
			double span_1 = actualtotal * .01;
			double inc_1 = span_1 / 33.0;
			double span_2 = actualtotal * .09;
			double inc_2 = span_2 / 33.0;
			double span_3 = actualtotal * .90;
			double inc_3 = span_3 / 33.0;
			double scale = 240000.0;
			scale += (inc_1 * (double)Math.min(slider, 33));
			slider -= Math.min(slider, 33);
			scale += (inc_2 * (double)Math.min(slider, 33));
			slider -= Math.min(slider, 33);
			scale += (inc_3 * (double)slider);
			long disptimespan = (long)scale;
			
			minDate = getFirstDisplayTime();
			maxDate = minDate + disptimespan;
			minDate -= getEndPadding();
			maxDate += getEndPadding();
		}

		/** Redraws the graphical representation of the stream.
		 */
		public void redraw( ) {
			
			// Relabel stuff in case language changed.
			zoomLabel.setText(Strings.get(141));
			autoZoom.setText(Strings.get(142));
			
			// Resize the image if the panel size has changed.
			int width = getWidth( );
			int height = getHeight( );

			// Quit if the panel is not visible yet.
			if (width <= 0 || height <= 0) { return; }
		
			// Prepare to draw to the double buffer.
			streamimage = 
				new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
			Graphics2D g = streamimage.createGraphics( );
			g.setColor( getBackground( ) );
			g.fillRect( 0, 0, width, height );

			// Some metrics.
			Color cardcolor = ( Color.WHITE );
			Color annotationrulecolor = new Color( 210, 210, 255 );
			Color reviewendtimecolor = new Color( 128, 64, 0 );
			Color currenttimecolor = new Color( 80, 40, 0 );
			Color highlightcolor = ( new Color(255, 0, 0 ) );
			Color lastcardcolor = ( new Color(128, 64, 64 ) );
			Color textcolor = ( Color.BLACK );
			int cardwidth = (int) (width * 0.4 );
			int cardxpos = (int) (width * 0.34);
			int timelabelxpos = (int)(width * 0.81);
			//g.setFont( FontController.getSizedFont(12) );
			g.setFont(g.getFont().deriveFont(12));
			
			// If there are no cards, or no stream, there's nothing to draw.
			if ( (stream == null) || stream.getCardCount() < 1 ) { 
				
				// Notify the user to add cards.
				int labelypos = (height/2)-40;
				g.setColor(textcolor);
				g.drawString( Strings.get(228), cardxpos-40, labelypos );
				labelypos+=13;
				g.drawString( Strings.get(229), cardxpos-40, labelypos );
				labelypos+=13;
				g.drawString( Strings.get(230), cardxpos-40, labelypos );
				
				this.repaint( );
				
				return; 
			}
	
			// Calculate important times in seconds.
			long currenttime = getCurrentTime() / 1000;
			long reviewendtime = getReviewEndTime() / 1000;

			// Cap the display range.
			long disptimespan = maxDate - minDate;
			if (minDate < (getFirstDisplayTime( ) - getEndPadding( ))) {
				minDate = getFirstDisplayTime( )- getEndPadding( );
				maxDate = minDate + disptimespan; // Keep previous range.
				adjustZoomSliderToRange();
			}
			if (maxDate > (getLastDisplayTime( ) + getEndPadding( ))) {
				maxDate = getLastDisplayTime( ) + getEndPadding( );
				adjustZoomSliderToRange();
			}

			// Get the display start and end times in seconds.
			earliestdtime = minDate / 1000;
			latestdtime = maxDate / 1000;
			
			// Calculate the desired display span and center.
			disptimespan = latestdtime - earliestdtime;
			
			// Calculate the height of one second in pixels.
			secondheight = (double)height / (double)disptimespan;
	
			// Draw the time label ruler lines and labels.
			String dtunits = Strings.get(130);
			long dtunitspan = 60;
			DecimalFormat df = new DecimalFormat("#0.0");
			df.setMaximumFractionDigits(1);
			if ( disptimespan < (2*3600) ) { // < 2 hours
				dtunits = Strings.get(130);
				df = new DecimalFormat("#0");
				dtunitspan = 60;
			} else if ( disptimespan < (2*86400) ) { // < 48 hours
				dtunits = Strings.get(131);
				dtunitspan = 3600;
			} else if ( disptimespan < (2*2592000) ) { // < 2 months
				dtunits = Strings.get(132);
				dtunitspan = 86400;
			} else if ( disptimespan < (4*2592000) ) { // < 4 months
				dtunits = Strings.get(133);
				dtunitspan = 604800;			
			} else if ( disptimespan < (4*77760000) ) { // < 2 years
				dtunits = Strings.get(134);
				dtunitspan = 2592000;
			} else {
				dtunits = Strings.get(135);
				dtunitspan = 77760000;
			}
			double disptimeincrementsecs = (double)disptimespan / 6.0;
			double disptimeincrement = 
				secondheight * (double)disptimeincrementsecs;
			// Mark time labels below the centeredtime.
			double curlabelypos = (double)currenttime - (double)earliestdtime;
			curlabelypos = height - ((double)(secondheight * curlabelypos));
			double curlabeltime = 0.0;
			curlabelypos += disptimeincrement;
			curlabeltime -= disptimeincrementsecs;
			long lowerlimit = height - 10;
			// Scan down to display start in case the display is way below the
			// current time.
			while (curlabelypos < 10) {
				curlabelypos += disptimeincrement;
				curlabeltime -= disptimeincrementsecs;
			}
			while (	curlabelypos < lowerlimit ) {
				double labeltime = ((double)curlabeltime / (double)dtunitspan);
				String lbl = df.format(labeltime) + " " + dtunits;
				long curlabeltick = (long)((curlabeltime+currenttime)*1000.0);
				int ccount = stream.getCardCountBefore( curlabeltick );
				g.setColor(annotationrulecolor);
				g.drawRect(0, (int)curlabelypos, width, 1);
				g.setColor(textcolor);
				g.drawString( lbl, timelabelxpos, (int)curlabelypos );
				g.drawString( String.valueOf(ccount)+Strings.get(136), 
						timelabelxpos, (int)curlabelypos+10 );
				curlabelypos += disptimeincrement;
				curlabeltime -= disptimeincrementsecs;
			}
			// Mark time labels above the centeredtime.
			curlabelypos = currenttime - earliestdtime;
			curlabelypos = height - ((int)(secondheight * curlabelypos));
			curlabeltime = 0;
			curlabelypos -= disptimeincrement;
			curlabeltime += disptimeincrementsecs;
			while ( curlabelypos > 10 ) {
				double labeltime = ((double)curlabeltime / (double)dtunitspan);
				String lbl = df.format(labeltime) + " " + dtunits;
				long curlabeltick = (long)((curlabeltime+currenttime)*1000.0);
				int ccount = stream.getCardCountBefore( curlabeltick );
				g.setColor(annotationrulecolor);
				g.drawRect(0, (int)curlabelypos, width, 1);
				g.setColor(textcolor);
				g.drawString( lbl, timelabelxpos, (int)curlabelypos );
				g.drawString( String.valueOf(ccount)+Strings.get(136), 
						timelabelxpos, (int)curlabelypos+10 );
				curlabelypos -= disptimeincrement;
				curlabeltime += disptimeincrementsecs;
			}

			// Draw the current time and review end time markers.
			long relativectime = currenttime - earliestdtime;
			int currenttimeypos = height - ((int)(secondheight*relativectime));
			long relativeretime = reviewendtime - earliestdtime;
			int reviewendtimeypos = 
				height-((int)(secondheight*relativeretime));
			g.setColor(currenttimecolor);
			drawdottedline( g, currenttimeypos );
			// But don't draw review end time if too close to current time.
			if (Math.abs(reviewendtimeypos-currenttimeypos) > 10) {
				g.setColor(reviewendtimecolor);
				drawdottedline( g, reviewendtimeypos );
			}
			
			// Draw every card, starting at the bottom of the panel and moving 
			// up.
			Iterator i = stream.iterator();
			g.setColor( cardcolor );
			while ( i.hasNext() ) {
				StreamCard sc = (StreamCard)i.next();
				long thiscardtime = 
					StreamCard.parseTime(sc.getBestNextReviewDate()) / 1000;
				thiscardtime -= earliestdtime; // Make relative to border.
				int thiscardypos = height - ((int)(secondheight*thiscardtime));
				if ( thiscardypos < getHeight() ) {
					if ( thiscardypos < 0 ) {
						// Finish if there are no more cards in display range.
						break;
					}
					g.fillRect( cardxpos, thiscardypos, cardwidth, 1 );
				}
			}

			// Color current card and last reviewed card if user is reviewing.
			StreamCard.StreamCardStats scs = stream.getLastCardStats();
			long curcardtime = stream.getCurrentCardBestReviewDate();
			long lastcardtime = (scs == null) ? -1 : scs.newReviewTime;
			// Highlight current card.
			int curcardypos = -1;
			int lastrevcardpos = -1;
			if ( curcardtime != -1 ) {
				g.setColor(highlightcolor);
				curcardtime /= 1000;
				curcardtime -= earliestdtime;
				curcardypos = height - ((int)(secondheight * curcardtime));
				g.fillRect( cardxpos, curcardypos, cardwidth, 1 );
			}
			// Highlight last reviewed card.
			if ( lastcardtime != -1 ) {
				g.setColor(lastcardcolor);
				lastcardtime /= 1000;
				lastcardtime -= earliestdtime;
				lastrevcardpos = height - ((int)(secondheight*lastcardtime));
				g.fillRect( cardxpos, lastrevcardpos, cardwidth, 1 );
				int lastrevcardlabelpos = lastrevcardpos-2;
				if ( lastrevcardlabelpos < 12 ) {
					lastrevcardlabelpos = lastrevcardpos + 10;
				}
				if ( lastrevcardlabelpos > 10 ) {
					g.drawString(Strings.get(137),(int)(width*0.20),
							lastrevcardlabelpos);
				}
			}
			// Don't label current if the label would overlap last reviewed:
			if ((lastcardtime == -1) || 
					(Math.abs(lastrevcardpos-curcardypos) > 10)) {
				g.drawString( 
						Strings.get(138), (int)(width * 0.20), curcardypos-2 );
			}
			
			// Color selected cards if user is browsing and has cards selected.
			if (stream.getCurrentCard() == null && selectedCards != null) {
				Iterator ii = selectedCards.iterator();
				while (ii.hasNext()) {
					StreamCard sc = new StreamCard( (IXMLElement)ii.next() );
					g.setColor(highlightcolor);
					long t = StreamCard.parseTime(sc.getBestNextReviewDate());
					t /= 1000;
					t -= earliestdtime;
					curcardypos = height - ((int)(secondheight * t));
					g.fillRect( cardxpos, curcardypos, cardwidth, 1 );					
				}
			}
	
			// Draw current and review time labels.
			g.setColor(currenttimecolor);
			g.drawString( 
					Strings.get(139), (int)(width * 0.75), currenttimeypos-2 );
			// But don't draw review end time if too close to current time.
			if (Math.abs(reviewendtimeypos-currenttimeypos) > 10) {
				g.setColor(reviewendtimecolor);
				int reviewendlabelypos = reviewendtimeypos - 2;
				// Flip label at top:
				if ( reviewendlabelypos < 12 ) {reviewendlabelypos+=13;}
				if (reviewendlabelypos > 10) {
					g.drawString(
						Strings.get(140),(int)(width*0.7),reviewendlabelypos);
				}
			}
		
			this.repaint( );
	
		} // end redraw
		
		private void drawdottedline( Graphics2D g, int ypos ) {
			int imax = getWidth( );
			for ( int x = 0; x < imax; x+=4 ) {
				g.drawRect(x, ypos, 1, 1);
			}
		}

		/** Called by the frame when the user has requested to move a group
		  * of cards.
		 */
		public void initiateMove() {
			if ((stream.getCurrentCard() == null) && 
					(selectedCards != null) && 
					(selectedCards.size()>0)) {
				isUserMovingCards = true;
				ownerframe.setMessage(Strings.get(207));
			}
		}

		public void onMoveDestinationSelected( int x, int y ) {
			
			int cardxpos = (int) (getWidth() * 0.34);
			int cardwidth = (int) (getWidth() * 0.4 );

			if (((x < cardxpos-10)&&(x < cardxpos-10)) ||
					((x > cardxpos+cardwidth+10)&&
							(x > cardxpos+cardwidth+10))) {
					
				// Not a valid location.  Cancel.
				selectedCards = null;
				ownerframe.setMessage(Strings.get(208));
				
			} else {
				
				ownerframe.onMoveSelectedCardsTo(yPosToTime(y), selectedCards);
				ownerframe.setMessage(Strings.get(209));
			}
			
			isUserMovingCards = false;
		}

		// Selects the card with the given ID.
		public void selectCardWithID( int id ) {
			StreamCard sc = stream.getFirstCardWithID(id);
			selectedCards = new Vector();
			selectedCards.add( sc.toXML() );		
		}
		
		/** Called when the user releases the mouse on the display panel.
		  * Tests whether the user clicked on a card and if so selects that
		  * card area unless a review is underway.
		  * @param x x position the mouse was clicked at. Relative to this panel.
		  * @param y y position the mouse was clicked at. Relative to this panel.
		 */
		public void onClick( int x, int y ) {
			
			// If this is not a review session, select or deselect and 
			// highlight according to the click.
			if ( stream.getCurrentCard() == null ) {

				if (selectionStart == null) {selectionStart = new Point(x,y);}
				if (selectionEnd == null) {selectionEnd = new Point(x,y);}
				
				Point p = new Point(
						Math.min(selectionStart.x, selectionEnd.x),
						Math.min(selectionStart.y, selectionEnd.y));
				Point q = new Point(
						Math.max(selectionStart.x, selectionEnd.x),
						Math.max(selectionStart.y, selectionEnd.y));
				
				// If the mouse was outside the card column, don't select.
				int cardxpos = (int) (getWidth() * 0.34);
				int cardwidth = (int) (getWidth() * 0.4 );
				if (((p.x < cardxpos-10)&&(q.x < cardxpos-10)) ||
					((p.x > cardxpos+cardwidth+10)&&
							(q.x > cardxpos+cardwidth+10))) {
					
					selectedCards = null;
					
				} else {
				
					// Calculate the time and card range that
					// was selected by the mouse drag.
					long endtime = yPosToTime(p.y-1);
					long starttime = yPosToTime(q.y+1);

					int firstcardindex = 
						stream.getFirstCardIndexAfterTime(starttime);
					int lastcardindex =
						stream.getFirstCardIndexBeforeTime(endtime);

					if ((firstcardindex > lastcardindex) ||
							(firstcardindex == -1) || (lastcardindex == -1)) {
						
						// No cards selected.
						selectedCards = null;

					} else {

						// Add the collection of selected cards.
						selectedCards = new Vector();
						Iterator it = stream.iterator();
						int i = 0;
						// Scan to first selected card.
						while ( it.hasNext() && i < firstcardindex ) {
							it.next();
							i++;
						}
						int total = 0;
						while ( it.hasNext() && i <= lastcardindex 
								&& total < 100 ) {
							StreamCard sc = (StreamCard)it.next();
							selectedCards.add(sc.toXML());
							i++;
							total++;
						}
						
						if (total == 100 && i <= lastcardindex) {
							ownerframe.setMessage(Strings.get(202));
						}

					}
				}

				selectionStart = null;
				selectionEnd = null;
				
				redraw();

				if ( ownerframe != null ) {
					ownerframe.onSelectionChanged();
				}
			}			
		}
		
		/** Converts a y position (in pixels) on the stream display panel to
		  * the corresponding time represented (in milliseconds).
		 */
		public long yPosToTime( int yPos ) {
						
			long seconds = earliestdtime;
			
			double pixelseconds = 1.0 / secondheight;
			seconds += (long)(((double)getHeight()-(double)yPos)*pixelseconds);
			
			return seconds*1000;			
		}
	
		// TODO: some kind of get and set selected card range mechanismm
		/** Returns the id of the selected card, or -1 if no card is selected.
		 */
		public int getSelectedCardID() {
			if (stream.getCurrentCard() != null || // Not in review
					selectedCards == null ||
					selectedCards.size() < 1 ) {
				return -1;
			} else {
				StreamCard sc = 
					new StreamCard((IXMLElement)selectedCards.firstElement());
				return sc.getID();
			}
		}
		
		/** Gets a set of cards around and including the selected card to 
		  * browse, or null if no cards are selected.
		 */
		public Vector getCardRange() {
			if ( stream.getCurrentCard() != null || // Not in review
					selectedCards == null ||
					selectedCards.size() < 1 ) {
				return null;
			} else {
				return selectedCards;
			}
		}
		
		public void setCardContainer( CardContainer cc ) {
			setStream((Stream)cc);
		}
	
		/** Gives this renderer an interface to access its card container.
		 */	
		public void setStream( Stream newstream ) {		
	
			stream = newstream;

			minDate = stream.getFrontCardBestReviewDate() - 7200000;
			maxDate = stream.getBackCardBestReviewDate() + 7200000;

			adjustZoomSliderToRange();

		} // end setStream
		
	} // end class StreamRenderPanel
	
	/** Panel for showing last reviewed card statistics:
	 */
	private class LastCardPanel extends JPanel {
		
		private Stream stream;
		private StreamCard.StreamCardStats scs;
		
		private BufferedImage img; // Double buffer.

		private Random random = new Random(new Date().getTime());
		
		// For flash animations on big hits and misses:
		private FlashTimerListener flashTimerListener = 
			new FlashTimerListener();
		private Timer flashTimer = new Timer( 50, flashTimerListener );
		private Color regularColor = getBackground( );
		private Color flashColor = Color.BLACK;
		private int animationFrame = 0;
		private String cheer, boo;

		private long longtime = 1209600000; // 2 weeks in millis

		public LastCardPanel( ) {
			setPreferredSize( new Dimension( 320, 80 ) );
			setBorder( BorderFactory.createTitledBorder( "" ) );
		}
		
		public void redraw(
				StreamCard.StreamCardStats new_scs, Stream newstream) {
			
			stream = newstream;

			cheer = 
				(new String[] {
						Strings.get(143),Strings.get(144),Strings.get(145)
							})[random.nextInt(3)];
			boo = 
				(new String[] {
						Strings.get(146),Strings.get(147),Strings.get(148)
							})[random.nextInt(3)];
			
			scs = new_scs;
			_redraw( );

			// Start flashing panel animation if necessary.
			if ( (Math.abs(scs.nextReviewSpanLong - 
					scs.lastReviewSpanLong) > longtime)
					|| (scs.nextReviewSpanLong == 
						stream.getMaxReviewTimeMillis())) {
				// Flash the panel:
				if (scs.nextReviewSpanLong == stream.getMinReviewTimeMillis())
				{
					// Flash Red:
					flashColor = Color.RED;
				} else {
					// Flash yellow:
					flashColor = Color.YELLOW;
				}
				animationFrame = 8;
				flashTimer.start();
			}
		}
		
		private void _redraw( ) {
			// Resize the image if the panel size has changed.
			int width = getWidth( );
			int height = getHeight( );

			// Quit if the panel is not visible yet.
			if (width <= 0 || height <= 0) { return; }
		
			// Prepare to draw to the double buffer.
			img = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
			Graphics2D g = img.createGraphics( );
			g.setColor( getBackground( ) );
			g.fillRect( 0, 0, width, height );
			g.setColor(Color.BLACK);
			//Font f = FontController.getSizedFont(16);
			//g.setFont( f );
			Font f = g.getFont().deriveFont(16);
			g.setFont(f);
			
			// Set up metrics.
			int y_increment = (height / 4) - 1;
			int indent = 10;
			int cur_x = 5;
			int cur_y = y_increment - 2;
			boolean bold = false;
			
			// Set the mouseover text.
			String s = Strings.get(149) + 
				StreamCard.formatRelativeShortTime(scs.lastReviewTime) +
				Strings.get(150) +
				StreamCard.formatRelativeShortTime(scs.newReviewTime);
			setToolTipText(s);

			// Draw the labels.
			s = Strings.get(151);
			if ( (scs.lastReviewSpanLong > scs.nextReviewSpanLong) ||
			 (scs.nextReviewSpanLong==stream.getMinReviewTimeMillis())) {
				s = Strings.get(152);
			}
			if ( ((scs.nextReviewSpanLong - scs.lastReviewSpanLong) > longtime)
				|| (scs.nextReviewSpanLong == 
					stream.getMaxReviewTimeMillis())) {
				s = Strings.get(153) + cheer + "):";
				bold = true;
			} else if ( (scs.nextReviewSpanLong - scs.lastReviewSpanLong) <
					(-1*longtime) ) {
				s = Strings.get(153) + boo + "):";
				bold = true;
			}
			if ( bold ) {g.setFont(f.deriveFont(Font.BOLD));}
			g.drawString(s, cur_x, cur_y);
			g.setFont(f);
			cur_y+=y_increment;
			cur_x+=indent;

			g.drawString(Strings.get(154) + scs.record, cur_x, cur_y);
			cur_y+=y_increment;

			s = Strings.get(155) + scs.lastReviewSpan + 
				Strings.get(156) + scs.nextReviewSpan;
			if ( bold ) {g.setFont(f.deriveFont(Font.BOLD));}
			g.drawString(s, cur_x, cur_y);
			g.setFont(f);
			cur_y+=y_increment;
			
			s = "";
			if (scs.nextReviewSpanLong == stream.getMaxReviewTimeMillis()) {
				if (scs.lastReviewSpanLong==stream.getMaxReviewTimeMillis()) {
					s += Strings.get(157);
				} else {
					s += Strings.get(158);
				}
			}
			if (scs.randomOffset != 0) {
				String absval = StreamCard.formatTimeSpanShortForm(
						Math.abs(scs.randomOffset));
				if (scs.randomOffset > 0) {s += "+"+absval;}
				else {s += "-"+absval;}
				if ( s.length() < 28 ) {s += Strings.get(159);}
			}
			g.drawString(s, cur_x, cur_y);
		}

		private class FlashTimerListener implements ActionListener {
			public void actionPerformed(ActionEvent evt) {
				animationFrame--;
				if (animationFrame % 2 == 0) {
					LastCardPanel.this.setBackground(regularColor);
					LastCardPanel.this._redraw();
				} else {
					LastCardPanel.this.setBackground(flashColor);
					LastCardPanel.this._redraw();
				}
				if ( animationFrame <= 0 ) {
					flashTimer.stop();
				}
			}
		};

		/** Paints this panel.
		 */
		public void paintComponent( Graphics g ) {
	
			super.paintComponent( g ); 
	
			if ( img != null  ) {
				g.drawImage( img, 0, 0, null );
			}
		}
		
	} // end class LastCardPanel
	
} // end class StreamDisplayPanel